/**
  *  \file client/scr-sim.cc
  *  \brief Battle Simulator Screen
  *
  *  This contains all the widgets and routines for editing a
  *  simulation.
  *
  *  (c) 2001-2012 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cctype>
#include "client/scr-sim.h"

#include "client/cscreen.h"
#include "client/dlg-fcode.h"
#include "client/scr-result.h"
#include "cpluslib/format.h"
#include "cpluslib/holder.h"
#include "cpluslib/substr.h"
#include "game/config.h"
#include "game/formula.h"
#include "game/spec.h"
#include "gfx/gfxcompl.h"
#include "gfx/palette.h"
#include "io/glob.h"
#include "ui/combobox.h"
#include "ui/dialogs/filedlg.h"
#include "ui/layout.h"
#include "ui/listbox.h"
#include "ui/msgbox.h"
#include "ui/table.h"
#include "ui/util.h"
#include "ui/window.h"
#include "util/except.h"
#include "util/misc.h"
#include "util/nls.h"
#include "defs.h"
#include "resmgr/resid.h"
#include "game/racename.h"
#include "game/host.h"
#include "ui/widgets/multilinestatic.h"
#include "ui/widgets/framegroup.h"
#include "ui/widgets/focusiterator.h"
#include "ui/widgets/statictext.h"
#include "resmgr/resmgr.h"
#include "ui/widgets/decimalsel.h"
#include "ui/widgets/inputline.h"
#include "ui/widgets/button.h"
#include "client/dialogs/fleetcost.h"
#include "client/widgets/helpwidget.h"
#include "client/dialogs/helpdlg.h"
#include "client/chart.h"
#include "game/team.h"

static GSimulatorGameInterface* sim_game_if;

/** List of battle participants, with editing functions. */
class WSimListWithHandler : public WSimList {
 public:
    WSimListWithHandler(GSimState& state);

    bool handleEvent(const UIEvent& e, bool second_pass);

    GSimShip* getCurrentShip();

    void addShip();
    void replicateShip();
    void deleteShip();

    void writeToGame();
    void updateFromGame();
    void switchToGame();

    void editAggressiveness();
    void editType(bool after_add);
    void editFCode();
    void incrFCode();
    void editPrimaryWeapon();
    void editSecondaryWeapon();
    void editName();
    void editId();
    void editDamage();
    void editShield();
    void editMass();
    void editCrew();
    void editExperienceLevel();
    void editOwner();
    void editEngine();
    void editInterceptId();
    void editAbilities();

    void swapUp();
    void swapDown();

    void toggleCloak();
    void toggleRandomFC();
    void toggleEnabled();

    void replicateCurrentShip(int count);
};

class WSimPlanetButton : public UIAbstractButton {
    GSimState& state;
    SignalConnection conn_edit;
    SignalConnection conn_structure;
 public:
    WSimPlanetButton(GSimState& state);
    ~WSimPlanetButton();
    void drawContent(GfxCanvas& can);
    void getLayoutInfo(LayoutInfo& info);
};

static string_t
getPlanetName(const GSimState& st)
{
    if (st.hasPlanet()) {
        string_t name;
        if (sim_game_if != 0)
            name = sim_game_if->getPlanetName(st.getPlanet().getId());
        else
            name = "?";
        if (st.getPlanet().getBaseBeamTech() > 0) {
            return format(_("%s (Planet+SB #%d)"), name, st.getPlanet().getId());
        } else {
            return format(_("%s (Planet #%d)"), name, st.getPlanet().getId());
        }
    } else {
        return _("none");
    }
}

static void
setPlanetName(GSimPlanet& pl)
{
    if (sim_game_if != 0) {
        pl.setName(sim_game_if->getPlanetName(pl.getId()));
    }
}

string_t
getShipName(const GSimShip& sh)
{
    const char* fmt;
    if (sh.getFlags() & sh.fl_Deactivated)
        fmt = _(" %d: %s (%s %s, disabled)");
    else
        fmt = _(" %d: %s (%s %s)");

    string_t hull_name;
    if (sh.isCustomShip())
        hull_name = _("custom ship");
    else
        hull_name = getHull(sh.getHull()).getShortName();

    return format(fmt, sh.getId(), sh.getName(), player_racenames.getAdjName(sh.getOwner()), hull_name);
}

static bool
sortById(const GSimShip& a, const GSimShip& b)
{
    return a.getId() < b.getId();
}

static bool
sortByOwner(const GSimShip& a, const GSimShip& b)
{
    if (a.getOwner() == b.getOwner())
        return a.getId() < b.getId();
    else
        return a.getOwner() < b.getOwner();
}

static bool
sortByHull(const GSimShip& a, const GSimShip& b)
{
    if (a.getHull() == b.getHull())
        return a.getId() < b.getId();
    else
        return a.getHull() < b.getHull();
}

static bool
sortByBattleOrder(const GSimShip& a, const GSimShip& b)
{
    if (WSimScreen::sim_opts.vcr_mode == GSimOptions::VcrHost || WSimScreen::sim_opts.vcr_mode == GSimOptions::VcrNuHost) {
        return sortByBattleOrderTHost(&a, &b);
    } else {
        /* sortByBattleOrderPHost takes non-const because it wants to cast them
           around to support planets. It still calls only non-modifying operations. */
        return sortByBattleOrderPHost(const_cast<GSimShip*>(&a), const_cast<GSimShip*>(&b));
    }
}

bool
WSimScreen::isExpGame()
{
    /* This used to check also the host mode. However, it was never implemented
       consistently. Therefore, enable editing experience whenever it is configured,
       even if it's not used in the current simulation. */
    return (config.NumExperienceLevels() > 0);
}

/** Graphically edit owner of an object. */
static bool
editOwner(GSimObject& obj)
{
    UIStringList list;
    for (int i = 1; i <= NUM_OWNERS; ++i)
        list.add(i, format("%X - %s", i, player_racenames.getShortName(i)));

    UIStandardListbox box(0, list.getCount(), font_ems[FONT_NORMAL] * 15, list);
    box.scrollToKey(obj.getOwner());
    if (box.doStandardListbox(_("Set Owner"))) {
        obj.setOwner(box.getKey());
        return true;
    } else {
        return false;
    }
}

/** Graphically edit experience level of an object. */
static void
editExperienceLevel(GSimObject& obj)
{
    if (config.NumExperienceLevels() > 0) {
        UIStringList list;
        for (int i = 0; i <= config.NumExperienceLevels(); ++i)
            list.add(i, format("%X - %s", i, getExperienceLevelName(i)));

        UIStandardListbox box(0, list.getCount(), font_ems[FONT_NORMAL] * 20, list);
        box.scrollToKey(obj.getExperienceLevel());
        if (box.doStandardListbox(_("Set Experience Level")))
            obj.setExperienceLevel(box.getKey());
    }
}

/******************************** WSimList *******************************/

static void
setColor(GfxContext& ctx, const GSimObject& obj)
{
    if ((obj.getFlags() & GSimObject::fl_Deactivated) != 0) {
        ctx.setColor(UIColor::tc_Faded);
    } else if (::getPlayerId() != 0) {
        // FIXME: be careful that we don't accidentally pull in the whole map renderer!
        ctx.setColor(getPlayerColor(obj.getOwner()));
    } else {
        // no change, keep default color
    }
}

/** Constructor.
    \param state       [in] Object to display/edit
    \param with_planet [in] Select whether planet is included in list */
WSimList::WSimList(GSimState& state, bool with_planet)
    : UIListbox(0, 20, 235,
                state.getNumShips() + ((with_planet && state.hasPlanet()) ? 1 : 0)),
      state(state),
      with_planet(with_planet),
      conn_ship_changed(state.sig_ship_changed.add(this, &WSimList::onShipChanged)),
      conn_planet_changed(state.sig_planet_changed.add(this, &WSimList::onPlanetChanged)),
      conn_structure_changed(state.sig_structure_changed.add(this, &WSimList::onStructureChanged))
{
    updateCount();
}

/** Destructor. */
WSimList::~WSimList()
{ }

void
WSimList::drawPart(GfxCanvas& can, int from, int to)
{
    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useFont(*fonts[FONT_NORMAL]);
    for (int i = from; i < to; ++i) {
        const unsigned int index = getTop() + i;
        GfxRect area;
        prepareEntry(ctx, ColorStyle, i, &area);
        if (index < state.getNumShips()) {
            /* ship entry */
            const GSimShip& sh = state.getShip(index);
            setColor(ctx, sh);
            outTextF(ctx, area.x, area.y, area.w, getShipName(sh));
        } else if (index == state.getNumShips() && with_planet && state.hasPlanet()) {
            /* planet entry */
            setColor(ctx, state.getPlanet());
            outTextF(ctx, area.x, area.y, area.w, " " + getPlanetName(state));
        }
    }
}

void
WSimList::onShipChanged(int /*index*/)
{
    drawWidget();
}

void
WSimList::onPlanetChanged()
{
    drawWidget();
}

void
WSimList::onStructureChanged()
{
    updateCount();
    drawWidget();
}

void
WSimList::updateCount()
{
    if (with_planet && state.hasPlanet())
        setSize(state.getNumShips() + 1);
    else
        setSize(state.getNumShips());
    adjustPosition();
}

/************************** WSimListWithHandler **************************/

/** Constructor.
    \param state [in] Simulation state to edit. */
WSimListWithHandler::WSimListWithHandler(GSimState& state)
    : WSimList(state, false)
{ }

bool
WSimListWithHandler::handleEvent(const UIEvent& e, bool second_pass)
{
    if (e.type == ev_Key) {
        switch (e.key.code) {
         case SDLK_INSERT:
            addShip();
            getSimState().doUpdates();
            return true;
         case SDLK_INSERT + ss_Ctrl:
         case '*':
            replicateShip();
            getSimState().doUpdates();
            return true;
         case SDLK_DELETE:
            deleteShip();
            getSimState().doUpdates();
            return true;
         case SDLK_F1:
            switchToGame();
            return true;
         case SDLK_UP + ss_Ctrl:
            swapUp();
            getSimState().doUpdates();
            return true;
         case SDLK_DOWN + ss_Ctrl:
            swapDown();
            getSimState().doUpdates();
            return true;
         case 'u':
            updateFromGame();
            getSimState().doUpdates();
            return true;
         case 'w':
            writeToGame();
            return true;
         case 'a':
            editAggressiveness();
            getSimState().doUpdates();
            return true;
         case 't':
            editType(false);
            getSimState().doUpdates();
            return true;
         case 'f':
            editFCode();
            getSimState().doUpdates();
            return true;
         case 'F':
            incrFCode();
            getSimState().doUpdates();
            return true;
         case '1':
            editPrimaryWeapon();
            getSimState().doUpdates();
            return true;
         case '2':
            editSecondaryWeapon();
            getSimState().doUpdates();
            return true;
         case 'n':
            editName();
            getSimState().doUpdates();
            return true;
         case 'i':
            editId();
            getSimState().doUpdates();
            return true;
         case 'd':
            editDamage();
            getSimState().doUpdates();
            return true;
         case 's':
            editShield();
            getSimState().doUpdates();
            return true;
         case 'm':
            editMass();
            getSimState().doUpdates();
            return true;
         case 'c':
            editCrew();
            getSimState().doUpdates();
            return true;
         case 'l':
            editExperienceLevel();
            getSimState().doUpdates();
            return true;
         case 'o':
            editOwner();
            getSimState().doUpdates();
            return true;
         case 'e':
            editEngine();
            getSimState().doUpdates();
            return true;
         case 'v':
            toggleCloak();
            getSimState().doUpdates();
            return true;
         case 'r':
            toggleRandomFC();
            getSimState().doUpdates();
            return true;
         case '.':
            toggleEnabled();
            getSimState().doUpdates();
            return true;
         case 'x':
            editInterceptId();
            getSimState().doUpdates();
            return true;
         case 'b':
            editAbilities();
            getSimState().doUpdates();
            return true;
        }
    }
    return WSimList::handleEvent(e, second_pass);
}

/** Get currently-selected ship.
    \return ship, or 0 if none selected */
GSimShip*
WSimListWithHandler::getCurrentShip()
{
    if (getSimState().getNumShips() == 0 || getCurrent() >= getSimState().getNumShips())
        return 0;
    return &getSimState().getShip(getCurrent());
}

/** Add new ship. */
void
WSimListWithHandler::addShip()
{
    GSimState& st = getSimState();
    if (st.getNumShips() >= st.MAX_SHIPS) {
        messageBox(_("This simulation already contains the maximum possible number of ships."),
                   _("Battle Simulator"));
    } else {
        if (st.getNumShips() == 0) {
            st.addShip();
            st.getShip(0).setHull(1);
            st.getShip(0).setId(1);
            st.getShip(0).setAggressiveness(GSimShip::agg_Kill);
            st.getShip(0).setName(_("Ship 1"));
            st.doUpdates();
            scrollTo(0, 1);
        } else {
            replicateCurrentShip(1);
            st.doUpdates();
        }

        editType(true);
    }
}

/** Replicate ship. User is asked for how often to replicate it. */
void
WSimListWithHandler::replicateShip()
{
    GSimState& st = getSimState();
    if (st.getNumShips() == 0) {
        // nothing there to replicate
    } else if (st.getNumShips() >= st.MAX_SHIPS) {
        messageBox(_("This simulation already contains the maximum possible number of ships."),
                   _("Battle Simulator"));
    } else {
        UIDecimalSelector count(1001, 0, st.MAX_SHIPS - st.getNumShips(), 1);
        if (count.doStandardDialog(_("Replicate Ship"), _("Number:"), 0 /* FIXME: help id */))
            replicateCurrentShip(count.getValue());
    }
}

/** Delete current ship. */
void
WSimListWithHandler::deleteShip()
{
    GSimState& st = getSimState();
    if (st.getNumShips() != 0) {
        st.removeShip(getCurrent());
        updateCount(); // do this here so it happens before updating the tiles
    }
}

/** Write this ship back to game. */
void
WSimListWithHandler::writeToGame()
{
    if (GSimShip* sh = getCurrentShip()) {
        if (sim_game_if->getShipRelation(*sh) == GSimulatorGameInterface::ShipIsPlayed) {
            /* only try ships that are played; for others, the 'w'
               button is disabled */
            if (!sim_game_if->updateToGame(*sh))
                messageBox(_("Ship status could not be written back."),
                           _("Battle Simulator"));
        }
    }
}

/** Update this ship from the game. */
void
WSimListWithHandler::updateFromGame()
{
    if (GSimShip* sh = getCurrentShip()) {
        if (sim_game_if->getShipRelation(*sh) != GSimulatorGameInterface::ShipIsUnknown) {
            /* only try ships that we recognize; for others, the 'u'
               button is disabled. In addition, updateFromGame() will
               update everything as long as the Id matches. */
            sim_game_if->updateFromGame(*sh);
        }
    }
}

/** Switch to game. */
void
WSimListWithHandler::switchToGame()
{
    if (GSimShip* sh = getCurrentShip()) {
        if (sim_game_if->selectGameShip(*sh)) {
            stop(1);
        }
    }
}

/** Edit aggressiveness. */
void
WSimListWithHandler::editAggressiveness()
{
    if (GSimShip* sh = getCurrentShip()) {
        UIStringList list;
        list.add(GSimShip::agg_Kill, _("! - Kill Mission"));
        list.add(GSimShip::agg_Passive, _("0 - Passive Ship"));
        for (int i = 1; i <= NUM_OWNERS; ++i)
            list.add(i, format(_("%X - Primary Enemy %s"), i, player_racenames.getAdjName(i)));
        list.add(GSimShip::agg_NoFuel, _("Z - No Fuel"));

        UIStandardListbox box(0, list.getCount(), 20 * font_ems[FONT_NORMAL], list);
        box.scrollToKey(sh->getAggressiveness());
        if (box.doStandardListbox(_("Set Aggressiveness"))) {
            int new_val = box.getKey();
            sh->setAggressiveness(new_val);
            if (new_val == GSimShip::agg_Kill || new_val == GSimShip::agg_NoFuel) {
                sh->setFlags(sh->getFlags() & ~GSimShip::fl_Cloaked);
                sh->setInterceptId(0);
            }
        }
    }
}

/** Edit current ship's hull type.

    \param after_add true iff this is the invocation of this dialog
    immediately after adding a ship. In this case, also tries to set a
    default owner for the ship. */
void
WSimListWithHandler::editType(bool after_add)
{
    if (GSimShip* sh = getCurrentShip()) {
        // FIXME: PCC 1.x had some features here which are still missing
        // (filtering, sorting)
        UIStringList list;
        for (int i = 1; i <= getNumHulls(); ++i)
            list.add(i, getHull(i).getName());
        list.sortAlphabetically();
        list.addAtFront(0, _("Custom Ship"));

        UIStandardListbox box(0, 20, getTextWidth(*fonts[FONT_NORMAL], list) + 20, list);
        box.scrollToKey(sh->getHull());
        if (box.doStandardListbox(_("Set Hull Type"))) {
            int new_val = box.getKey();
            int old_val = sh->getHull();
            sh->setHull(new_val);

            if (new_val != old_val && new_val != 0 && after_add) {
                // pick default owner if possible
                int defo = 0;
                bool we_can_build_it = false;
                for (int pl = 1; pl <= NUM_PLAYERS; ++pl) {
                    if (getTruehullSlot(pl, new_val) != 0) {
                        if (defo == 0)
                            defo = pl;
                        else
                            defo = -1;
                        if (pl == sh->getOwner())
                            we_can_build_it = true;
                    }
                }
                if (defo > 0) {
                    if (defo != sh->getOwner()) {
                        /* ship has default owner */
                        sh->setOwner(defo);
                        sh->setFCode("?""?""?");
                        if (sh->getAggressiveness() == sh->getOwner()) {
                            sh->setAggressiveness(GSimShip::agg_Kill);
                        }
                    }
                } else {
                    if (!we_can_build_it) {
                        /* ship does not have a default owner, so we copy another ship's owner.
                           Do not copy the FCode, to avoid surprises */
                        sh->setFCode("?""?""?");
                    }
                }
            }
        }
    }
}

/** Edit current ship's friendly code. */
void
WSimListWithHandler::editFCode()
{
    if (GSimShip* sh = getCurrentShip()) {
        /* build list of available codes */
        GFCodeList filtered_list;
        GFCodeList& full_list = GFCodeList::getInstance();

        int our_flags = GFCode::fc_Ship;
        if (sh->getBayCount() != 0 || sh->getTorpLauncherCount() != 0 || sh->getBeamCount() != 0)
            our_flags |= GFCode::fc_Capital;
        // FIXME: alchemy, registered

        for (GFCodeList::iterator i = full_list.begin(); i != full_list.end(); ++i) {
            if ((i->getFlags() & GFCode::fc_Ship) == 0)
                continue;
            if ((i->getFlags() & GFCode::fc_Capital) != 0 && (our_flags & GFCode::fc_Capital) == 0)
                continue;
            filtered_list.addFCode(*i);
        }

        WFCodeWindow window(_("Set Friendly Code"), filtered_list);
        window.setFCode(sh->getFCode());

        ui_root->centerWidget(window);
        if (ui_root->runModal(window) == cm_Enter) {
            sh->setFCode(window.getFCode());

            // FIXME: PCC1 displays a help window to tell the user why it is doing this
            sh->assignRandomFCodeFlags();
        }
    }
}

/** Assign friendly code sequence. */
void
WSimListWithHandler::incrFCode()
{
    // Find current ship, scroll down, find new current ship
    GSimShip* prev = getCurrentShip();
    scrollTo(getCurrent()+1, 1);
    GSimShip* now = getCurrentShip();

    // If both exist and differ, perform our operation
    if (prev != 0 && now != 0 && prev != now) {
        // Copy friendly code and flags
        string_t fcode = prev->getFCode();
        int      flags = ((now->getFlags() & ~GSimShip::fl_RandomDigits)
                          | (prev->getFlags() & GSimShip::fl_RandomDigits));

        // Initialize fcode
        if (fcode.size() < 3)
            fcode.append(' ', 3 - fcode.size());
        for (int i = 0; i < 3; ++i) {
            if (((flags & GSimShip::fl_RandomFC) != 0
                 && (flags & (GSimShip::fl_RandomFC1 << i)) != 0)
                || (fcode[i] < '0' || fcode[i] > '9'))
            {
                fcode[i] = '0' + (rand() % 10);
            }
        }

        // Find new fcode. Attempt to make it unique, but don't
        // crash if that's not possible (because we've made the fcode
        // completely numeric above, that will not happen anyway, but
        // it might happen if someone makes a .ccb file with >1000
        // entries).
        for (int tries = 0; tries < NUM_SHIPS; ++tries) {
            // Increase fcode
            for (int i = 3; i > 0; --i) {
                ++fcode[i-1];
                if (fcode[i-1] <= '9')
                    break;
                fcode[i-1] = '0';
            }

            // Check for uniqueness
            bool ok = true;
            for (int i = 0; i < getSimState().getNumShips() && ok; ++i)
                if (i != getCurrent() && getSimState().getShip(i).getFCode() == fcode)
                    ok = false;
            if (ok)
                break;
        }

        // Assign new fcode
        now->setFCode(fcode);
        now->setFlags(flags);
    }
}

/** Edit current ship's primary weapon. */
void
WSimListWithHandler::editPrimaryWeapon()
{
    if (GSimShip* sh = getCurrentShip()) {
        int max_beams = 0;
        if (sh->isCustomShip()) {
            max_beams = 20; // FIXME: distinction Host/PHost
        } else {
            max_beams = getHull(sh->getHull()).getMaxBeams();
        }
        if (sh->getBeamCount() != 0 || max_beams != 0) {
            /* build combo box */
            UIStringList list;
            list.add(0, _("No weapon"));
            // FIXME: the drop-down list is missing specs and hot-keys
            for (int i = 1; i <= NUM_BEAMS; ++i)
                list.add(i, getBeam(i).getName());
            UIComboBox type_select(1000, 0, NUM_BEAMS, list);
            if (sh->getBeamCount() == 0)
                type_select.setValue(NUM_BEAMS);
            else
                type_select.setValue(sh->getBeamType());

            /* count */
            UIDecimalSelector count_select(1001, 0, max_beams, 1);
            count_select.setValue(sh->getBeamCount());

            /* build window */
            ObjectHolder h;
            UIWindow w(_("Set Primary Weapon"), UIWindow::BLUE_WINDOW, COLOR_GRAY);
            w.setLayout(UIVBoxLayout::instance5);

            UIGroup& controls = h.add(new UIGroup(h.add(new UIGridLayout(2))));
            controls.add(h.add(new UIStaticText(_("Type:"), UIStaticText::StaticTitle)));
            createComboBoxControls(controls, type_select, h);
            controls.add(h.add(new UIStaticText(_("Count:"), UIStaticText::StaticTitle)));
            controls.add(count_select); // FIXME: buttons

            UIGroup& buttons = h.add(new UIGroup(UIHBoxLayout::instance5));
            buttons.add(h.add(new UIButton(cm_Enter, _("OK"), SDLK_RETURN)).addStop());
            buttons.add(h.add(new UIButton(cm_Escape, _("Cancel"), SDLK_ESCAPE)).addStop());
            buttons.add(h.add(new UISpacer()));

            w.add(controls);
            w.add(buttons);
            w.add(h.add(new UIFocusIterator(1000, 1001, UIFocusIterator::fi_Vert)));
            w.pack();
            type_select.setFocusedWidget();

            ui_root->centerWidget(w);
            int rv = ui_root->runModal(w);

            if (rv == cm_Enter) {
                if (type_select.getValue() != 0 && count_select.getValue() != 0) {
                    sh->setBeamCount(count_select.getValue());
                    sh->setBeamType(type_select.getValue());
                } else {
                    sh->setBeamCount(0);
                    sh->setBeamType(0);
                }
            }
        }
    }
}

/** Edit current ship's secondary weapon. */
void
WSimListWithHandler::editSecondaryWeapon()
{
    if (GSimShip* sh = getCurrentShip()) {
        if (sh->isCustomShip()) {
            /* custom ship can have any combination */
            /* first build list of types */
            UIStringList list;
            list.add(0, _("No weapon"));
            for (int i = 1; i <= NUM_TORPS; ++i)
                // FIXME: the drop-down list is missing specs and hot-keys
                list.add(i, getTorp(i).getName());
            list.add(NUM_TORPS+1, _("Fighters"));

            UIComboBox type_select(1000, 0, NUM_TORPS+1, list);

            /* count */
            UIDecimalSelector count_select(1001, 0, 20 /* FIXME: PHost... */, 1);
            count_select.setValue(sh->getTorpLauncherCount());

            if (sh->getBayCount() != 0) {
                type_select.setValue(NUM_TORPS+1);
                count_select.setValue(sh->getBayCount());
            } else if (sh->getTorpLauncherCount() != 0) {
                type_select.setValue(sh->getTorpType());
                count_select.setValue(sh->getTorpLauncherCount());
            } else {
                type_select.setValue(NUM_TORPS);
                count_select.setValue(0);
            }

            /* ammo */
            UIDecimalSelector ammo_select(1002, 0, 10000, 10);
            ammo_select.setValue(sh->getAmmo());

            /* build window */
            ObjectHolder h;
            UIWindow w(_("Set Secondary Weapon"), UIWindow::BLUE_WINDOW, COLOR_GRAY);
            w.setLayout(UIVBoxLayout::instance5);

            UIGroup& controls = h.add(new UIGroup(h.add(new UIGridLayout(2))));
            controls.add(h.add(new UIStaticText(_("Type:"), UIStaticText::StaticTitle)));
            createComboBoxControls(controls, type_select, h);
            controls.add(h.add(new UIStaticText(_("Count:"), UIStaticText::StaticTitle)));
            controls.add(count_select); // FIXME: buttons
            controls.add(h.add(new UIStaticText(_("Ammo:"), UIStaticText::StaticTitle)));
            controls.add(ammo_select);

            UIGroup& buttons = h.add(new UIGroup(UIHBoxLayout::instance5));
            buttons.add(h.add(new UIButton(cm_Enter, _("OK"), SDLK_RETURN)).addStop());
            buttons.add(h.add(new UIButton(cm_Escape, _("Cancel"), SDLK_ESCAPE)).addStop());
            buttons.add(h.add(new UISpacer()));

            w.add(controls);
            w.add(buttons);
            w.add(h.add(new UIFocusIterator(1000, 1002, UIFocusIterator::fi_Vert)));
            w.pack();
            type_select.setFocusedWidget();

            ui_root->centerWidget(w);
            int rv = ui_root->runModal(w);

            if (rv == cm_Enter) {
                sh->setAmmo(ammo_select.getValue());
                if (count_select.getValue() != 0 && type_select.getValue() != 0) {
                    if (type_select.getValue() == NUM_TORPS+1) {
                        sh->setTorpLauncherCount(0);
                        sh->setTorpType(0);
                        sh->setBayCount(count_select.getValue());
                    } else {
                        sh->setTorpLauncherCount(count_select.getValue());
                        sh->setTorpType(type_select.getValue());
                        sh->setBayCount(0);
                    }
                } else {
                    sh->setTorpLauncherCount(0);
                    sh->setTorpType(0);
                    sh->setBayCount(0);
                }
            }
        } else {
            int max_tubes = getHull(sh->getHull()).getMaxLaunchers();
            int num_bays = getHull(sh->getHull()).getNumBays();
            int max_ammo = getHull(sh->getHull()).getCargo();
            if (num_bays != 0) {
                /* it's a carrier; just ask for number of fighters. */
                sh->setBayCount(num_bays);
                sh->setTorpType(0);
                sh->setTorpLauncherCount(0);

                UIDecimalSelector select(0, 0, max_ammo, 10);
                select.setValue(sh->getAmmo());
                if (select.doStandardDialog(_("Set Number of Fighters"), _("Fighters:"), 0))
                    sh->setAmmo(select.getValue());
            } else if (max_tubes != 0) {
                /* it's a torper, so we have to do a more complicated
                   dialog. */
                sh->setBayCount(0);

                /* first build list of types */
                UIStringList list;
                list.add(0, _("No weapon"));
                for (int i = 1; i <= NUM_TORPS; ++i)
                    list.add(i, getTorp(i).getName());

                UIComboBox type_select(1000, 0, NUM_TORPS, list);
                if (sh->getTorpLauncherCount() == 0)
                    type_select.setValue(NUM_TORPS);
                else
                    type_select.setValue(sh->getTorpType());

                /* count */
                UIDecimalSelector count_select(1001, 0, max_tubes, 1);
                count_select.setValue(sh->getTorpLauncherCount());

                /* ammo */
                UIDecimalSelector ammo_select(1002, 0, max_ammo, 10);
                ammo_select.setValue(sh->getAmmo());

                /* build window */
                ObjectHolder h;
                UIWindow w(_("Set Secondary Weapon"), UIWindow::BLUE_WINDOW, COLOR_GRAY);
                w.setLayout(UIVBoxLayout::instance5);

                UIGroup& controls = h.add(new UIGroup(h.add(new UIGridLayout(2))));
                controls.add(h.add(new UIStaticText(_("Type:"), UIStaticText::StaticTitle)));
                createComboBoxControls(controls, type_select, h);
                controls.add(h.add(new UIStaticText(_("Count:"), UIStaticText::StaticTitle)));
                controls.add(count_select); // FIXME: buttons
                controls.add(h.add(new UIStaticText(_("Torpedoes:"), UIStaticText::StaticTitle)));
                controls.add(ammo_select);

                UIGroup& buttons = h.add(new UIGroup(UIHBoxLayout::instance5));
                buttons.add(h.add(new UIButton(cm_Enter, _("OK"), SDLK_RETURN)).addStop());
                buttons.add(h.add(new UIButton(cm_Escape, _("Cancel"), SDLK_ESCAPE)).addStop());
                buttons.add(h.add(new UISpacer()));

                w.add(controls);
                w.add(buttons);
                w.add(h.add(new UIFocusIterator(1000, 1002, UIFocusIterator::fi_Vert)));
                w.pack();
                type_select.setFocusedWidget();

                ui_root->centerWidget(w);
                int rv = ui_root->runModal(w);

                if (rv == cm_Enter) {
                    sh->setAmmo(ammo_select.getValue());
                    if (count_select.getValue() != 0 && type_select.getValue() != 0) {
                        sh->setTorpLauncherCount(count_select.getValue());
                        sh->setTorpType(type_select.getValue());
                    } else {
                        sh->setTorpLauncherCount(0);
                        sh->setTorpType(0);
                    }
                }
            }
        }
    }
}

/** Edit current ship's name. */
void
WSimListWithHandler::editName()
{
    if (GSimShip* sh = getCurrentShip()) {
        string_t s = sh->getName();
        if (UIInputLine::run(s, 20, _("Set Ship Name"), _("Name:"), 20, UIInputLine::ilf_GameChars))
            sh->setName(s);
    }
}

/** Edit current ship's Id number. */
void
WSimListWithHandler::editId()
{
    if (GSimShip* sh = getCurrentShip()) {
        UIDecimalSelector select(0, 1, NUM_SHIPS, 10);
        select.setValue(sh->getId());
        while (1) {
            if (!select.doStandardDialog(_("Set Id"), _("Id:"), 0))
                break;

            bool id_is_in_use = false;
            for (GSimState::ship_index_t i = 0; i < getSimState().getNumShips(); ++i) {
                GSimShip& ish = getSimState().getShip(i);
                if (&ish != sh && ish.getId() == select.getValue()) {
                    id_is_in_use = true;
                    break;
                }
            }

            if (!id_is_in_use) {
                sh->setId(select.getValue());
                break;
            }

            messageBox(_("This Id number is already in use in this simulation setup. Please choose another one."),
                       _("Battle Simulator"));
        }
    }
}

/** Edit current ship's damage level. */
void
WSimListWithHandler::editDamage()
{
    if (GSimShip* sh = getCurrentShip()) {
        UIDecimalSelector select(0, 0, sh->getMaxDamage(), 10);
        select.setValue(sh->getDamage());
        if (select.doStandardDialog(_("Set Damage"), _("Damage [%]:"), 0)) {
            int val = select.getValue();
            sh->setDamage(val);
            if (sh->getShield() > sh->getMaxShield())
                sh->setShield(sh->getMaxShield());
        }
    }
}

/** Edit current ship's shield level. */
void
WSimListWithHandler::editShield()
{
    if (GSimShip* sh = getCurrentShip()) {
        UIDecimalSelector select(0, 0, sh->getMaxShield(), 10);
        select.setValue(sh->getShield());
        if (select.doStandardDialog(_("Set Shield Level"), _("Shield [%]:"), 0))
            sh->setShield(select.getValue());
    }
}

/** Edit current ship's mass. This only works on custom ships. */
void
WSimListWithHandler::editMass()
{
    if (GSimShip* sh = getCurrentShip()) {
        if (sh->isCustomShip()) {
            UIDecimalSelector select(0, 1, 10000, 100);
            select.setValue(sh->getMass());
            if (select.doStandardDialog(_("Set Mass"), _("Mass [kt]:"), 0))
                sh->setMass(select.getValue());
        }
    }
}

/** Edit current ship's crew amount. */
void
WSimListWithHandler::editCrew()
{
    if (GSimShip* sh = getCurrentShip()) {
        int limit;
        if (sh->isCustomShip())
            limit = 10000;
        else
            limit = getHull(sh->getHull()).getCrew();

        UIDecimalSelector select(0, 1, limit, 10);
        select.setValue(sh->getCrew());
        if (select.doStandardDialog(_("Set Crew"), _("Crew:"), 0))
            sh->setCrew(select.getValue());
    }
}

/** Edit current ship's experience level. */
void
WSimListWithHandler::editExperienceLevel()
{
    if (GSimShip* sh = getCurrentShip())
        ::editExperienceLevel(*sh);
}

/** Edit current ship's owner. */
void
WSimListWithHandler::editOwner()
{
    if (GSimShip* sh = getCurrentShip())
        ::editOwner(*sh);
}

/** Edit current ship's engine type. */
void
WSimListWithHandler::editEngine()
{
    if (GSimShip* sh = getCurrentShip()) {
        UIStringList list;
        for (int i = 1; i <= NUM_ENGINES; ++i)
            list.add(i, format("%X - %s", i, getEngine(i).getName()));

        UIStandardListbox box(0, list.getCount(), font_ems[FONT_NORMAL] * 20, list);
        box.scrollToKey(sh->getEngineType());
        if (box.doStandardListbox(_("Set Engine Type")))
            sh->setEngineType(box.getKey());
    }
}

/** Edit current ship's intercept target. */
void
WSimListWithHandler::editInterceptId()
{
    if (GSimShip* sh = getCurrentShip()) {
        /* FIXME: use a better widget */
        UIDecimalSelector select(0, 0, NUM_SHIPS, 10);
        select.setValue(sh->getInterceptId());
        if (select.doStandardDialog(_("Set Intercept-Attack Target"), _("Id:"), 0)) {
            sh->setInterceptId(select.getValue());
            if (sh->getInterceptId() != 0) {
                /* cannot cloak and intercept */
                sh->setFlags(sh->getFlags() & ~GSimShip::fl_Cloaked);
                if (const GSimShip* target = getSimState().getShipById(sh->getInterceptId())) {
                    /* pick correct PE. FIXME: if we have a standing
                       'enemies' order, and a nonzero PE, we don't
                       need to change here. */
                    sh->setAggressiveness(target->getOwner());
                } else if (int n = sim_game_if->getShipOwner(sh->getInterceptId())) {
                    /* we're intercepting a ship which is not in the
                       simulation, but in the game. Set correct PE in
                       case ship is added later */
                    sh->setAggressiveness(n);
                } else if (sh->getAggressiveness() == GSimShip::agg_Kill || sh->getAggressiveness() == GSimShip::agg_NoFuel) {
                    /* cannot intercept and kill, cannot intercept
                       without fuel. */
                    sh->setAggressiveness(0);
                }
            }
        }
    }
}

/** Toggle cloak status on current ship. */
void
WSimListWithHandler::toggleCloak()
{
    if (GSimShip* sh = getCurrentShip()) {
        int flags = sh->getFlags();
        flags ^= GSimShip::fl_Cloaked;
        if (flags & GSimShip::fl_Cloaked) {
            /* cannot cloak and be aggressive */
            if (sh->getAggressiveness() == GSimShip::agg_Kill || sh->getAggressiveness() == GSimShip::agg_NoFuel)
                sh->setAggressiveness(GSimShip::agg_Passive);
            /* cannot cloak and intercept */
            sh->setInterceptId(0);
        }
        sh->setFlags(flags);
    }
}

/** Toggle "random FCode" function. */
void
WSimListWithHandler::toggleRandomFC()
{
    if (GSimShip* sh = getCurrentShip()) {
        sh->setFlags(sh->getFlags() ^ GSimShip::fl_RandomFC);
    }
}

/** Toggle whether thip is enabled. */
void
WSimListWithHandler::toggleEnabled()
{
    if (GSimShip* sh = getCurrentShip()) {
        sh->setFlags(sh->getFlags() ^ GSimShip::fl_Deactivated);
    }
}

/** Replicate current ship. This is the back-end function. Caller must
    make sure that the maximum possible number of ships is not
    exceeded if he wants to notify user. This function will silently
    ignore failures. Failure here means inability to obtain a free Id
    number, which only happens when there are more than NUM_SHIPS
    ships.

    \param count number of times to replicate */
void
WSimListWithHandler::replicateCurrentShip(int count)
{
    GSimState& st = getSimState();

    // Build bitset of used ship Ids. The naive version of this routine,
    // using getFreeId() to obtain new ship Ids, is O(n^3), taking a
    // noticeable amount of time (>1s at 2 GHz) when inserting 998 ships
    // (=the maximum).
    uint8 used_id_bitset[NUM_SHIPS/8 + 1];
    zeroFill(used_id_bitset);
    for (GSimState::ship_index_t i = 0; i < st.getNumShips(); ++i) {
        int id = st.getShip(i).getId();
        if (id > 0 && id <= NUM_SHIPS)
            used_id_bitset[id >> 3] |= (1 << (id & 7));
    }

    // Go
    int pos = getCurrent();
    int id = 0;
    for (int i = 0; i < count; ++i) {
        do
            ++id;
        while (id <= NUM_SHIPS && (used_id_bitset[id >> 3] & (1 << (id & 7))) != 0);
        if (id > NUM_SHIPS)
            break;
        used_id_bitset[id >> 3] |= (1 << (id & 7));
        st.duplicateShip(pos, id);
        ++pos;
    }
    getSimState().doUpdates();
    scrollTo(pos, 1);
}

/**************************** WSimPlanetButton ***************************/

/** Constructor.
    \param state State to observe */
WSimPlanetButton::WSimPlanetButton(GSimState& state)
    : UIAbstractButton(UIKey('p')),
      state(state),
      conn_edit(observe(state.sig_planet_changed)),
      conn_structure(observe(state.sig_structure_changed))
{ }

WSimPlanetButton::~WSimPlanetButton()
{ }

void
WSimPlanetButton::drawContent(GfxCanvas& can)
{
    GfxRect r = getExtent();

    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(GfxStandardColorScheme::instance);

    drawFrameDown(can, r);
    r.grow(-1, -1);
    drawSolidBar(ctx, r, COLOR_GRAY);
    r.grow(-4, -3);

    ctx.useFont(*fonts[FONT_NORMAL]);
    if (state.hasPlanet()) {
        ctx.setColor(COLOR_BLACK);
        outTextF(ctx, r.x, r.y, r.w, getPlanetName(state));
    } else {
        ctx.setColor(COLOR_DARK);
        outTextF(ctx, r.x, r.y, r.w, _("(no planet or base)"));
    }
    can.update(getExtent());
}

void
WSimPlanetButton::getLayoutInfo(LayoutInfo& info)
{
    info.min_size = font_sizes[FONT_NORMAL].scaledBy(10, 1).movedBy(0, 10);
    info.pref_size = font_sizes[FONT_NORMAL].scaledBy(18, 1).movedBy(0, 10);
    info.flags = li_FlexibleH;
}

/********************************* Tiles *********************************/

/** Simulator ship information tile.

    Displays information about the current ship in a simulation. This
    class provides the necessary interface to WSimList; descendants
    only need to override the display methods. */
class WSimShipTile : public WTile {
 protected:
    WSimList& list;
 private:
    SignalConnection conn_moved, conn_ship_changed, conn_structure_changed;

 protected:
    WSimShipTile(WSimList& list, int width, int height);
    ~WSimShipTile();

    void onChange();
    void onShipEdit(int slot);
};

WSimShipTile::WSimShipTile(WSimList& list, int width, int height)
    : WTile(GfxPoint(width, height)), list(list),
      conn_moved(list.sig_moved.add(this, &WSimShipTile::onChange)),
      conn_ship_changed(list.getSimState().sig_ship_changed.add(this, &WSimShipTile::onShipEdit)),
      conn_structure_changed(list.getSimState().sig_structure_changed.add(this, &WSimShipTile::onChange))
{ }

WSimShipTile::~WSimShipTile()
{ }

void
WSimShipTile::onShipEdit(int slot)
{
    if (slot == list.getCurrent())
        drawWidget();
}

void
WSimShipTile::onChange()
{
    drawWidget();
}

/**************************** WSimShipTypeTile ***************************/

class WSimShipTypeTile : public WSimShipTile {
    ObjectHolder h;
 public:
    WSimShipTypeTile(WSimList& list);
    void drawData(GfxCanvas& can);
    void getLayoutInfo(LayoutInfo& info);
};

class WSimShipDamageTile : public WSimShipTile {
    void drawContent(UITableWriter& tw, bool with_data);
    ObjectHolder h;
 public:
    WSimShipDamageTile(WSimList& list);
    void drawData(GfxCanvas& can);
};

class WSimShipWeaponTile : public WSimShipTile {
    void drawContent(UITableWriter& tw, bool with_data);
    ObjectHolder h;
 public:
    WSimShipWeaponTile(WSimList& list);
    void drawData(GfxCanvas& can);
};

class WSimShipImageTile : public WSimShipTile {
 public:
    WSimShipImageTile(WSimList& list);
    void drawData(GfxCanvas& can);
};

/** Button on ship editor. Automatically manages its disabledness
    state depending on current ship. */
class WSimShipButton : public UIButton {
    WSimList& list;
    SignalConnection conn_moved, conn_ship_changed, conn_structure_changed;

 public:
    WSimShipButton(WSimList& list, const char* name, UIKey key);
    ~WSimShipButton();

    void onUpdate();
};

WSimShipButton::WSimShipButton(WSimList& list, const char* name, UIKey key)
    : UIButton(name, key),
      list(list),
      conn_moved(list.sig_moved.add(this, &WSimShipButton::onUpdate)),
      conn_ship_changed(list.getSimState().sig_ship_changed.add(this, &WSimShipButton::onUpdate)),
      conn_structure_changed(list.getSimState().sig_structure_changed.add(this, &WSimShipButton::onUpdate))
{
    if (name[0] != 0 && name[1] == 0 && name[0] != 'U' && name[0] != 'W')
        setState(bf_SmallSize, true);
    onUpdate();
}

WSimShipButton::~WSimShipButton()
{ }

void
WSimShipButton::onUpdate()
{
    bool enable;
    bool press = false;
    const GSimState& st = list.getSimState();
    if (st.getNumShips() == 0) {
        /* no ships => all controls disabled */
        enable = false;
    } else {
        const GSimShip& this_ship = list.getSimState().getShip(list.getCurrent());
        switch (getKey()) {
         case SDLK_RETURN:
            enable = (st.getNumShips() > 1 || st.hasPlanet());
            break;

         case '1':
            enable = (this_ship.isCustomShip()
                      || getHull(this_ship.getHull()).getMaxBeams() != 0);
            break;

         case '2':
            enable = (this_ship.isCustomShip()
                      || getHull(this_ship.getHull()).getMaxLaunchers() != 0
                      || getHull(this_ship.getHull()).getNumBays() != 0);
            break;

         case SDLK_F1:
         case 'w':
            enable = (sim_game_if->getShipRelation(this_ship) == GSimulatorGameInterface::ShipIsPlayed);
            break;

         case 'u':
            enable = (sim_game_if->getShipRelation(this_ship) != GSimulatorGameInterface::ShipIsUnknown);
            break;

         case 'm':
            enable = this_ship.isCustomShip();
            break;

         case 'l':
            enable = WSimScreen::isExpGame();
            break;

         default:
            enable = true;
            break;
        }

        if (getKey() == 'r') {
            press = (this_ship.getFlags() & GSimShip::fl_RandomFC) != 0;
        }
    }

    setState(st_Disabled, !enable);
    setState(bf_Highlight, press);
}

WSimShipTypeTile::WSimShipTypeTile(WSimList& list)
    : WSimShipTile(list, 15 * font_ems[FONT_NORMAL], 6 * font_heights[FONT_NORMAL])
{
    // Translators: this button always is the 'r' key.
    const char* r = _("Random#");
    int em = font_heights[FONT_NORMAL];
    int em1 = em-1;
    int rwi = fonts[FONT_SMALL]->getTextWidth(r) + 8;
    int rwi5 = rwi+2;

    setLayout(RightJustified, 0, true);
    addControl(GfxRect(rwi5,    0, em1, em1), h.add(new WSimShipButton(list, "N", 'n')));
    addControl(GfxRect(rwi5,   em, em1, em1), h.add(new WSimShipButton(list, "T", 't')));
    addControl(GfxRect(rwi5, 2*em, em1, em1), h.add(new WSimShipButton(list, "O", 'o')));
    addControl(GfxRect(rwi5, 3*em, em1, em1), h.add(new WSimShipButton(list, "I", 'i')));
    addControl(GfxRect(rwi5, 4*em, em1, em1), h.add(new WSimShipButton(list, "F", 'f')));

    WSimShipButton& rb = h.add(new WSimShipButton(list, r, 'r'));
    rb.setState(rb.bf_SmallSize, true);
    addControl(GfxRect(0,    4*em, rwi, em1), rb);
}

void
WSimShipTypeTile::drawData(GfxCanvas& can)
{
    GfxContext ctx;
    ctx.useCanvas(can).useColorScheme(getSkin());
    UITableWriter tw(ctx, getInnerSize(0, font_heights[FONT_NORMAL]), FONT_NORMAL);

    tw.newLColumn();
    tw.writeFixedCell(UIColor::tc_Static, _("Name:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Type:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Owner:"));
    tw.writeFixedCell(UIColor::tc_Static, _("ID:"));
    tw.writeFixedCell(UIColor::tc_Static, _("FCode:"));
    // FIXME: rating

    tw.newLColumnUpTo(getExtent().w);
    const int index = list.getCurrent();
    if (index >= 0 && index < list.getSimState().getNumShips()) {
        const GSimShip& sh = list.getSimState().getShip(index);
        tw.writeVariableCell(UIColor::tc_Green, sh.getName());
        if (sh.isCustomShip())
            tw.writeVariableCell(UIColor::tc_Green, _("Custom Ship"));
        else
            tw.writeVariableCell(UIColor::tc_Green, getHull(sh.getHull()).getName());
        tw.writeVariableCell(UIColor::tc_Green, player_racenames.getShortName(sh.getOwner()));
        tw.writeVariableCell(UIColor::tc_Green, itoa(sh.getId()));

        // Friendly-code is special: use color-coding for randomized
        // digits. This cannot be done directly with UITableWriter.
        // Hence, remember current position and write a blank cell,
        // then write fcode manually.
        GfxPoint pos = tw.getCurrentPosition();
        tw.writeVariableCell(UIColor::tc_Green, string_t());
        int digits = 0;
        if (sh.getFlags() & sh.fl_RandomFC) {
            digits = (sh.getFlags() & sh.fl_RandomDigits);
            if (digits == 0)
                digits = sh.fl_RandomDigits;
        }
        string_t fc = sh.getFCode();
        for (string_t::size_type i = 0; i < 3 && i < fc.size(); ++i) {
            if (digits & (sh.fl_RandomFC1 << i))
                ctx.setColor(UIColor::tc_Yellow);
            else
                ctx.setColor(UIColor::tc_Green);
            outText(ctx, pos.x, pos.y, string_t(fc, i, 1));
            pos.x += fonts[FONT_NORMAL]->getTextWidth(string_t(fc, i, 1));
        }
    } else {
        for (int i = 0; i < 5; ++i)
            tw.writeVariableCell(UIColor::tc_Green, string_t());
    }
    tw.finish();
}

void
WSimShipTypeTile::getLayoutInfo(LayoutInfo& info)
{
    /* This tile sits on the same line as the WSimShipImageTile, and we want it
       to occupy as many horizontal space as possible. The WSimShipImageTile
       should not be scaled. */
    WSimShipTile::getLayoutInfo(info);
    info.flags |= li_FlexibleH;
}

WSimShipDamageTile::WSimShipDamageTile(WSimList& list)
    : WSimShipTile(list, 20 * font_ems[FONT_NORMAL], 3 * font_heights[FONT_NORMAL])
{
    const int em = font_heights[FONT_NORMAL];
    const int em1 = em-1;

    setLayout(RightJustified, 0, true);
    addControl(GfxRect(em,     0, em1, em1), h.add(new WSimShipButton(list, "D", 'd')));
    addControl(GfxRect(0,     em, em1, em1), h.add(new WSimShipButton(list, "L", 'l')));
    addControl(GfxRect(em,    em, em1, em1), h.add(new WSimShipButton(list, "C", 'c')));
    addControl(GfxRect(0,   2*em, em1, em1), h.add(new WSimShipButton(list, "M", 'm')));
    addControl(GfxRect(em,  2*em, em1, em1), h.add(new WSimShipButton(list, "S", 's')));
}

void
WSimShipDamageTile::drawContent(UITableWriter& tw, bool with_data)
{
    tw.newLColumn();
    tw.writeFixedCell(UIColor::tc_Static, _("Damage:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Crew:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Shield:"));

    if (with_data) {
        const int index = list.getCurrent();
        if (index >= 0 && index < list.getSimState().getNumShips()) {
            const GSimShip& sh = list.getSimState().getShip(index);

            tw.newRColumn(5 * font_ems[FONT_NORMAL]);
            tw.writeVariableCell(UIColor::tc_Green, numToString(sh.getDamage()));
            tw.writeVariableCell(UIColor::tc_Green, numToString(sh.getCrew()));
            tw.writeVariableCell(UIColor::tc_Green, numToString(sh.getShield()));

            tw.newLColumnUpTo(2 * font_ems[FONT_NORMAL]);
            tw.writeVariableCell(UIColor::tc_Green, "%");
            tw.writeVariableCell(UIColor::tc_Green, "");
            tw.writeVariableCell(UIColor::tc_Green, "%");

            tw.newLColumn();
            tw.writeVariableCell(UIColor::tc_Green, "");
            if (WSimScreen::isExpGame())
                tw.writeVariableCell(UIColor::tc_Green, getExperienceLevelName(sh.getExperienceLevel()));
            else
                tw.writeVariableCell(UIColor::tc_Green, "");
            // FIXME: turn this into a proper function of GSimShip
            int mass = sh.isCustomShip() ? sh.getMass() : ::getHull(sh.getHull()).getMass();
            tw.writeVariableCell(UIColor::tc_Green, format(_("%d kt"), mass));
        } else {
            /* FIXME */
        }
    }
    tw.finish();
}

void
WSimShipDamageTile::drawData(GfxCanvas& can)
{
    GfxContext ctx;
    ctx.useCanvas(can).useColorScheme(getSkin());
    UITableWriter tw(ctx, getInnerSize(0, font_heights[FONT_NORMAL]*2), FONT_NORMAL);
    drawContent(tw, true);
}



WSimShipWeaponTile::WSimShipWeaponTile(WSimList& list)
    // FIXME: original has additional space between lines
    : WSimShipTile(list, 25 * font_ems[FONT_NORMAL], 7 * font_heights[FONT_NORMAL])
{
    const int em = font_heights[FONT_NORMAL];
    const int em1 = em-1;

    setLayout(RightJustified, 0, true);
    addControl(GfxRect(em,         0, em1, em1), h.add(new WSimShipButton(list, "1", '1')));
    addControl(GfxRect(em,        em, em1, em1), h.add(new WSimShipButton(list, "2", '2')));
    addControl(GfxRect(em,      3*em, em1, em1), h.add(new WSimShipButton(list, "E", 'e')));
    addControl(GfxRect(em,      4*em, em1, em1), h.add(new WSimShipButton(list, "A", 'a')));
    addControl(GfxRect(0,       4*em, em1, em1), h.add(new WSimShipButton(list, "V", 'v')));
    addControl(GfxRect(em,      5*em, em1, em1), h.add(new WSimShipButton(list, "B", 'b')));
    addControl(GfxRect(em,      6*em, em1, em1), h.add(new WSimShipButton(list, "X", 'x')));
}

void
WSimShipWeaponTile::drawContent(UITableWriter& tw, bool with_data)
{
    tw.newLColumn();
    tw.writeFixedCell(UIColor::tc_Static, _("Primary:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Secondary:"));
    tw.writeFixedCell(UIColor::tc_Static, "");
    tw.writeFixedCell(UIColor::tc_Static, _("Engine:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Aggressive:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Abilities:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Intercept-Att.:"));

    if (with_data) {
        const int index = list.getCurrent();
        if (index >= 0 && index < list.getSimState().getNumShips()) {
            const GSimShip& sh = list.getSimState().getShip(index);

            tw.newLColumn();
            if (sh.getBeamCount() != 0)
                tw.writeVariableCell(UIColor::tc_Green, format(_("%d \xC3\x97 %s"), sh.getBeamCount(), getBeam(sh.getBeamType()).getName()));
            else
                tw.writeVariableCell(UIColor::tc_Green, _("none"));

            if (sh.getTorpLauncherCount() != 0) {
                tw.writeVariableCell(UIColor::tc_Green, format(_("%d %s launcher%!1{s%}"), sh.getTorpLauncherCount(), getTorp(sh.getTorpType()).getName()));
                tw.writeVariableCell(UIColor::tc_Green, format(_("%d torpedo%!1{es%}"), sh.getAmmo()));
            } else if (sh.getBayCount() != 0) {
                tw.writeVariableCell(UIColor::tc_Green, format(_("%d fighter bay%!1{s%}"), sh.getBayCount()));
                tw.writeVariableCell(UIColor::tc_Green, format(_("%d fighter%!1{s%}"), sh.getAmmo()));
            } else {
                tw.writeVariableCell(UIColor::tc_Green, _("none"));
                tw.writeVariableCell(UIColor::tc_Green, "");
            }

            tw.writeVariableCell(UIColor::tc_Green, getEngine(sh.getEngineType()).getName());

            string_t agg;
            if (sh.getAggressiveness() == sh.agg_Kill)
                agg = _("Kill");
            else if (sh.getAggressiveness() >= 1 && sh.getAggressiveness() <= NUM_OWNERS)
                agg = format("Enemy %s", player_racenames.getShortName(sh.getAggressiveness()));
            else if (sh.getAggressiveness() == sh.agg_NoFuel)
                agg = _("No fuel");
            else
                agg = _("Passive");
            if (sh.getFlags() & sh.fl_Cloaked)
                agg += _(", cloaked");
            tw.writeVariableCell(UIColor::tc_Green, agg);

            agg.clear();
            if (sh.hasShipFunction(sf_PlanetImmunity))
                agg += _("Planet Immunity"), agg += ", ";
            if (sh.hasShipFunction(sf_FullWeaponry))
                agg += _("Full Weaponry"), agg += ", ";
            if (sh.hasShipFunction(sf_Commander))
                agg += _("Commander"), agg += ", ";
            if (agg.size())
                agg.erase(agg.size() - 2);
            else
                agg = _("none");
            if (sh.hasAnyNonstandardFunction())
                tw.writeVariableCell(UIColor::tc_Yellow, agg);
            else
                tw.writeVariableCell(UIColor::tc_Green, agg);

            if (sh.getInterceptId() == 0)
                agg = _("none");
            else if (const GSimShip* other = list.getSimState().getShipById(sh.getInterceptId()))
                agg = format("%s (#%d)", other->getName(), other->getId());
            else
                agg = format(_("Ship #%d"), sh.getInterceptId());
            tw.writeVariableCell(UIColor::tc_Green, agg);
        } else {
            /* FIXME */
        }
    }
    tw.finish();
}

void
WSimShipWeaponTile::drawData(GfxCanvas& can)
{
    GfxContext ctx;
    ctx.useCanvas(can).useColorScheme(getSkin());
    UITableWriter tw(ctx, getInnerSize(0, font_heights[FONT_NORMAL]*2), FONT_NORMAL);
    drawContent(tw, true);
}


WSimShipImageTile::WSimShipImageTile(WSimList& list)
    : WSimShipTile(list, 105+2, 93+2)
{ }

void
WSimShipImageTile::drawData(GfxCanvas& can)
{
    GfxRect inner = getInnerSize(0, 0);
    drawSolidBar(can, inner, standard_colors[0]);
    drawFrameDown(can, inner);

    const int index = list.getCurrent();
    if (index >= 0 && index < list.getSimState().getNumShips()) {
        const int hullnr = list.getSimState().getShip(index).getHull();

        Ptr<GfxPixmap> pix;
        if (hullnr > 0)
            pix = requestPixmap(ResId(res::ship, getHull(hullnr).getPicture(), hullnr));
        else
            pix = requestPixmap(ResId(res::ship, 200, 0));

        if (pix) {
            GfxRect clip(0, 0,
                         std::min(inner.w - 2, pix->getWidth()),
                         std::min(inner.h - 2, pix->getHeight()));
            can.blitSurface(GfxPoint(1, 1), *pix, clip);
        }
    } else {
        // TRANSLATIONS: Must be two lines
        string_t text = _("Press [Ins] to\nadd a ship");

        GfxContext ctx;
        ctx.useCanvas(can);
        ctx.useColorScheme(GfxStandardColorScheme::instance);
        ctx.useFont(*fonts[FONT_SMALL]);
        ctx.setTextAlign(1, 1);
        ctx.setColor(COLOR_WHITE);

        GfxPoint pt = inner.getCenter();
        outTextFormatted(ctx, pt.x, pt.y-font_heights[FONT_SMALL]/2, extent.w, text);
    }
}

/**************************** Function Editor ****************************/

class WSimFunctionEditor;

class WSimFunctionTile : public WTile {
    WSimFunctionEditor& editor;
    void drawContent(UITableWriter& writer, bool with_content);
    ObjectHolder h;
 public:
    WSimFunctionTile(WSimFunctionEditor& editor);
    void drawData(GfxCanvas& can);
};

class WSimFunctionEditor : public UIWindow {
    GSimShip& ship;
    ObjectHolder holder;
    int new_flags;
    WTile* ptile;

    void toggle(GSimFunction sf);
 public:
    WSimFunctionEditor(GSimShip& sh);
    bool handleEvent(const UIEvent& e, bool second_pass);
    const char* getFunctionType(GSimFunction sf) const;
    int getNewFlags() const
        { return new_flags; }
};

class WSimFunctionButton : public UIButton {
 public:
    WSimFunctionButton(const char* name, UIKey key)
        : UIButton(name, key)
        { setState(bf_SmallSize, true); }
};

WSimFunctionTile::WSimFunctionTile(WSimFunctionEditor& editor)
    : WTile(font_sizes[FONT_NORMAL].scaledBy(24, 3)),
      editor(editor)
{
    const int em = font_heights[FONT_NORMAL];
    const int em1 = em-1;

    setLayout(RightJustified, 0, true);
    addControl(GfxRect(0,         0, em1, em1), h.add(new WSimFunctionButton("I", 'i')));
    addControl(GfxRect(0,        em, em1, em1), h.add(new WSimFunctionButton("F", 'f')));
    addControl(GfxRect(0,    2 * em, em1, em1), h.add(new WSimFunctionButton("C", 'c')));
}

void
WSimFunctionTile::drawContent(UITableWriter& writer, bool with_content)
{
    writer.newLColumnUpTo(font_ems[FONT_NORMAL]*11);
    writer.writeFixedCell(UIColor::tc_Static, _("Planet Immunity:"));
    writer.writeFixedCell(UIColor::tc_Static, _("Full Weaponry:"));
    writer.writeFixedCell(UIColor::tc_Static, _("Commander"));
    if (with_content) {
        writer.newRColumn(font_ems[FONT_NORMAL]*11);
        writer.writeVariableCell(UIColor::tc_Green, editor.getFunctionType(sf_PlanetImmunity));
        writer.writeVariableCell(UIColor::tc_Green, editor.getFunctionType(sf_FullWeaponry));
        writer.writeVariableCell(UIColor::tc_Green, editor.getFunctionType(sf_Commander));
    }
    writer.finish();
}

void
WSimFunctionTile::drawData(GfxCanvas& can)
{
    GfxContext ctx;
    ctx.useCanvas(can).useColorScheme(getSkin());
    UITableWriter tw(ctx, getInnerSize(0, font_heights[FONT_NORMAL]), FONT_NORMAL);
    drawContent(tw, true);
}

WSimFunctionEditor::WSimFunctionEditor(GSimShip& ship)
    : UIWindow(_("Ship Abilities"), BLUE_WINDOW, COLOR_GRAY),
      ship(ship),
      new_flags(ship.getFlags()),
      ptile(0)
{
    setLayout(UIVBoxLayout::instance5);

    UIGroup& buttons = holder.add(new UIGroup(UIHBoxLayout::instance5));

    WTile& tile = holder.add(new WSimFunctionTile(*this));

    buttons.add(holder.add(new UIButton(cm_Enter,  _("OK"), SDLK_RETURN)).addStop());
    buttons.add(holder.add(new UIButton(cm_Escape, _("Cancel"), SDLK_ESCAPE)).addStop());
    buttons.add(holder.add(new UISpacer()));
    // buttons.add(holder.add(new UIButton(cm_Help, _("H"), 'h')));

    add(tile);
    add(buttons);
    pack();

    // WSimPlanetEditor uses a signal for this; since we have just one client, we can
    // as well just keep a pointer.
    ptile = &tile;
}

void
WSimFunctionEditor::toggle(GSimFunction sf)
{
    const int value = GSimShip::simf_to_value_bit[sf];
    const int set = GSimShip::simf_to_set_bit[sf];
    const int flags = new_flags;
    if ((flags & set) == 0) {
        // default => Yes
        new_flags |= (value | set);
    } else if ((flags & value) != 0) {
        // Yes => No
        new_flags &= ~value;
    } else {
        // No => Default
        new_flags &= ~set;
    }

    if (ptile)
        ptile->drawWidget();
}

bool
WSimFunctionEditor::handleEvent(const UIEvent& e, bool second_pass)
{
    if (e.type == ev_Key) {
        if (e.key.code == 'i') {
            toggle(sf_PlanetImmunity);
            return true;
        } else if (e.key.code == 'f') {
            toggle(sf_FullWeaponry);
            return true;
        } else if (e.key.code == 'c') {
            toggle(sf_Commander);
            return true;
        }
    }
    return UIWindow::handleEvent(e, second_pass);
}

const char*
WSimFunctionEditor::getFunctionType(GSimFunction sf) const
{
    const int value = GSimShip::simf_to_value_bit[sf];
    const int set = GSimShip::simf_to_set_bit[sf];
    const int flags = new_flags;
    if ((flags & set) != 0) {
        if ((flags & value) != 0) {
            return _("yes");
        } else {
            return _("no");
        }
    } else {
        if (ship.hasImpliedHullFunction(sf)) {
            return _("default (=yes)");
        } else {
            return _("default (=no)");
        }
    }
}

// FIXME: position of these functions?

/** Edit current ship's abilities. */
void
WSimListWithHandler::editAbilities()
{
    if (GSimShip* sh = getCurrentShip()) {
        WSimFunctionEditor fedit(*sh);
        ui_root->centerWidget(fedit);
        ui_root->add(fedit);
        int result = ui_root->run();
        ui_root->close(fedit);
        if (result == cm_Enter)
            sh->setFlags(fedit.getNewFlags());
    }
}

/** Move current ship upwards by swapping with the ship above it. */
void
WSimListWithHandler::swapUp()
{
    int n = getCurrent();
    if (n > 0) {
        scrollTo(n-1, 1);
        getSimState().swapShips(n, n-1);
    }
}

/** Move current ship downwards by swapping with the ship below it. */
void
WSimListWithHandler::swapDown()
{
    int n = getCurrent();
    if (n + 1 < getSimState().getNumShips()) {
        scrollTo(n+1, 1);
        getSimState().swapShips(n, n+1);
    }
}

/****************************** Planet Tiles *****************************/

class WSimPlanetEditor;

class WSimPlanetPlanetTile : public WTile {
    // name N, owner O, id I, fcode F, defense posts PD, level L
    GSimPlanet& planet;
    ObjectHolder h;

    void drawContent(UITableWriter& writer, bool with_content);
 public:
    WSimPlanetPlanetTile(GSimPlanet& planet);

    void drawData(GfxCanvas& can);
};

class WSimPlanetBaseTile : public WTile {
    // beam tech B, fighters G, sb defense S, torps T, weapons W
    GSimPlanet& planet;
    ObjectHolder h;

    void drawContent(UITableWriter& writer, bool with_content);
 public:
    WSimPlanetBaseTile(GSimPlanet& planet);

    void drawData(GfxCanvas& can);
};

class WSimPlanetEditButton : public UIButton {
    GSimPlanet& planet;
 public:
    WSimPlanetEditButton(GSimPlanet& planet, const char* title, UIKey key);
    void update();
};

class WSimPlanetEditor : public UIWindow {
    GSimPlanet& planet;
    ObjectHolder holder;
    Signal0 sig_edit;

 public:
    WSimPlanetEditor(GSimPlanet& planet);

    void editName();
    void editOwner();
    void editId();
    void editFCode();
    void editDefense();
    void editPopulation();
    void editBaseBeamTech();
    void editBaseFighters();
    void editBaseDefense();
    void editBaseTorpTech();
    void editBaseTorps();

    bool handleEvent(const UIEvent& e, bool second_pass);
    void update();
};

WSimPlanetPlanetTile::WSimPlanetPlanetTile(GSimPlanet& planet)
    : WTile(font_sizes[FONT_NORMAL].scaledBy(30, WSimScreen::isExpGame() ? 6 : 5)),
      planet(planet)
{
    setHeading(_("Planet:"));
    setDisplayState(sNormal);

    const int em = font_heights[FONT_NORMAL];
    const int em1 = em-1;

    setLayout(RightJustified, 0, true);
    addControl(GfxRect(em,       0, em1, em1), h.add(new WSimPlanetEditButton(planet, "N", 'n')));
    addControl(GfxRect(em,      em, em1, em1), h.add(new WSimPlanetEditButton(planet, "O", 'o')));
    addControl(GfxRect(em,  2 * em, em1, em1), h.add(new WSimPlanetEditButton(planet, "I", 'i')));
    addControl(GfxRect(em,  3 * em, em1, em1), h.add(new WSimPlanetEditButton(planet, "F", 'f')));
    addControl(GfxRect(em,  4 * em, em1, em1), h.add(new WSimPlanetEditButton(planet, "D", 'd')));
    addControl(GfxRect(0,   4 * em, em1, em1), h.add(new WSimPlanetEditButton(planet, "P", 'p')));
    if (WSimScreen::isExpGame())
        addControl(GfxRect(em, 5 * em, em1, em1), h.add(new WSimPlanetEditButton(planet, "L", 'l')));
}

void
WSimPlanetPlanetTile::drawContent(UITableWriter& writer, bool with_content)
{
    writer.newLColumn();
    writer.writeFixedCell(UIColor::tc_Static, _("Name:"));
    writer.writeFixedCell(UIColor::tc_Static, _("Owner:"));
    writer.writeFixedCell(UIColor::tc_Static, _("ID:"));
    writer.writeFixedCell(UIColor::tc_Static, _("FCode:"));
    writer.writeFixedCell(UIColor::tc_Static, _("Defense Posts:"));
    if (WSimScreen::isExpGame())
        writer.writeFixedCell(UIColor::tc_Static, _("Level:"));

    if (with_content) {
        writer.newLColumn();
        if (sim_game_if != 0)
            writer.writeVariableCell(UIColor::tc_Green, sim_game_if->getPlanetName(planet.getId()));
        else
            writer.writeVariableCell(UIColor::tc_Green, "-");
        writer.writeVariableCell(UIColor::tc_Green, player_racenames.getShortName(planet.getOwner()));
        writer.writeVariableCell(UIColor::tc_Green, numToString(planet.getId()));
        writer.writeVariableCell(UIColor::tc_Green, planet.getFCode());
        writer.writeVariableCell(UIColor::tc_Green, numToString(planet.getDefense()));
        if (WSimScreen::isExpGame()) {
            writer.writeVariableCell(UIColor::tc_Green, getExperienceLevelName(planet.getExperienceLevel()));
        }
    }
    writer.finish();
}

void
WSimPlanetPlanetTile::drawData(GfxCanvas& can)
{
    GfxContext ctx;
    ctx.useCanvas(can).useColorScheme(getSkin());
    UITableWriter writer(ctx, getInnerSize(10, font_heights[FONT_NORMAL]*2), FONT_NORMAL);
    drawContent(writer, true);
}

WSimPlanetBaseTile::WSimPlanetBaseTile(GSimPlanet& planet)
    : WTile(font_sizes[FONT_NORMAL].scaledBy(30, 5)),
      planet(planet)
{
    setHeading(_("Starbase:"));
    setDisplayState(sNormal);

    const int em = font_heights[FONT_NORMAL];
    const int em1 = em-1;

    setLayout(RightJustified, 0, true);
    addControl(GfxRect(0,    0, em1, em1), h.add(new WSimPlanetEditButton(planet, "B", 'b')));
    addControl(GfxRect(0,   em, em1, em1), h.add(new WSimPlanetEditButton(planet, "G", 'g')));
    addControl(GfxRect(0, 2*em, em1, em1), h.add(new WSimPlanetEditButton(planet, "S", 's')));
    addControl(GfxRect(0, 3*em, em1, em1), h.add(new WSimPlanetEditButton(planet, "T", 't')));
    addControl(GfxRect(0, 4*em, em1, em1), h.add(new WSimPlanetEditButton(planet, "W", 'w')));
}

void
WSimPlanetBaseTile::drawContent(UITableWriter& writer, bool with_content)
{
    writer.newLColumn();
    writer.writeFixedCell(UIColor::tc_Static, _("Beam Tech:"));
    writer.writeFixedCell(UIColor::tc_Static, _("Fighters:"));
    writer.writeFixedCell(UIColor::tc_Static, _("SB Defense:"));
    writer.writeFixedCell(UIColor::tc_Static, _("Torp Tech:"));
    writer.writeFixedCell(UIColor::tc_Static, _("# Torpedoes:"));

    if (with_content) {
        writer.newLColumn();
        if (planet.getBaseBeamTech() != 0) {
            writer.writeVariableCell(UIColor::tc_Green, numToString(planet.getBaseBeamTech()));
            writer.writeVariableCell(UIColor::tc_Green, numToString(planet.getBaseFighters()));
            writer.writeVariableCell(UIColor::tc_Green, numToString(planet.getBaseDefense()));
            writer.writeVariableCell(UIColor::tc_Green, numToString(planet.getBaseTorpTech()));
            writer.writeVariableCell(UIColor::tc_Green, numToString(planet.getBaseTorpsAsType(planet.getBaseTorpTech())));
        } else {
            writer.writeVariableCell(UIColor::tc_Green, _("(no starbase)"));
            for (int i = 0; i < 4; ++i)
                writer.writeVariableCell(UIColor::tc_Green, string_t());
        }
    }
    writer.finish();
}

void
WSimPlanetBaseTile::drawData(GfxCanvas& can)
{
    GfxContext ctx;
    ctx.useCanvas(can).useColorScheme(getSkin());
    UITableWriter writer(ctx, getInnerSize(10, font_heights[FONT_NORMAL]), FONT_NORMAL);
    drawContent(writer, true);
}

WSimPlanetEditButton::WSimPlanetEditButton(GSimPlanet& planet, const char* title, UIKey key)
    : UIButton(title, key), planet(planet)
{
    setState(bf_SmallSize, true);
    update();
}

void
WSimPlanetEditButton::update()
{
    bool enabled;
    switch (getKey()) {
     case 'g':
     case 's':
     case 't':
     case 'w':
        enabled = (planet.getBaseBeamTech() != 0);
        break;
     default:
        enabled = true;
        break;
    }
    setState(st_Disabled, !enabled);
}


WSimPlanetEditor::WSimPlanetEditor(GSimPlanet& planet)
    : UIWindow(_("Edit Planet"), BLUE_WINDOW, COLOR_GRAY),
      planet(planet)
{
    setLayout(UIVBoxLayout::instance5);

    UIGroup& buttons = holder.add(new UIGroup(UIHBoxLayout::instance5));

    UIBaseWidget& tile1 = holder.add(new WSimPlanetPlanetTile(planet));
    UIBaseWidget& tile2 = holder.add(new WSimPlanetBaseTile(planet));

    tile1.observe(sig_edit);
    tile2.observe(sig_edit);

    buttons.add(holder.add(new UIButton(cm_Enter,  _("OK"), SDLK_RETURN)).addStop());
    buttons.add(holder.add(new UIButton(cm_Escape, _("Cancel"), SDLK_ESCAPE)).addStop());
    buttons.add(holder.add(new UIButton(cm_No,     _("Del - No Planet"), SDLK_DELETE)).addStop());
    buttons.add(holder.add(new UISpacer()));
    buttons.add(holder.add(new UIButton(_("H"), 'h')));

    add(tile1);
    add(tile2);
    add(buttons);
    pack();

    /* q & d */
    for (UIBaseWidget* w = tile1.getFirstChild()->getFirstChild(); w != 0; w = w->getNextSibling())
        if (WSimPlanetEditButton* e = dynamic_cast<WSimPlanetEditButton*>(w))
            sig_edit.add(e, &WSimPlanetEditButton::update);
    for (UIBaseWidget* w = tile2.getFirstChild()->getFirstChild(); w != 0; w = w->getNextSibling())
        if (WSimPlanetEditButton* e = dynamic_cast<WSimPlanetEditButton*>(w))
            sig_edit.add(e, &WSimPlanetEditButton::update);
}

bool
WSimPlanetEditor::handleEvent(const UIEvent& e, bool second_pass)
{
    if (e.type == ev_Key) {
        switch (e.key.code) {
         case 'n':
            editName();
            update();
            return true;
         case 'o':
            editOwner();
            update();
            return true;
         case 'i':
            editId();
            update();
            return true;
         case 'f':
            editFCode();
            update();
            return true;
         case 'd':
            editDefense();
            update();
            return true;
         case 'p':
            editPopulation();
            update();
            return true;
         case 'l':
            ::editExperienceLevel(planet);
            update();
            break;
         case 'b':
            editBaseBeamTech();
            update();
            return true;
         case 'g':
            editBaseFighters();
            update();
            return true;
         case 's':
            editBaseDefense();
            update();
            return true;
         case 't':
            editBaseTorpTech();
            update();
            return true;
         case 'w':
            editBaseTorps();
            update();
            return true;
         case 'h':
         case ss_Alt + 'h':
         case SDLK_F1:
            doHelp("pcc2:simplanet");
            return true;
        }
    }
    return UIWindow::handleEvent(e, second_pass);
}

void
WSimPlanetEditor::update()
{
    sig_edit.raise();
}

/** Edit planet name. Since planet names can usually not be changed,
    we only offer a list of all names available in the game, and
    change the Id accordingly. */
void
WSimPlanetEditor::editName()
{
    if (!sim_game_if)
        return;

    UIStringList list;
    for (int i = 1; i <= NUM_PLANETS; ++i)
        list.add(i, sim_game_if->getPlanetName(i));
    list.sortAlphabetically();

    UIStandardListbox box(0, 20, 20 * font_ems[FONT_NORMAL], list);
    box.scrollToKey(planet.getId());
    if (box.doStandardListbox(_("Set Planet Name"), 0)) {
        planet.setId(box.getKey());
        setPlanetName(planet);
    }
}

/** Edit planet owner. */
void
WSimPlanetEditor::editOwner()
{
    if (::editOwner(planet)) {
        if (planet.getBaseBeamTech() != 0) {
            int limit = config.MaximumFightersOnBase(planet.getOwner());
            if (planet.getBaseFighters() > limit)
                planet.setBaseFighters(limit);
            limit = config.MaximumDefenseOnBase(planet.getOwner());
            if (planet.getBaseDefense() > limit)
                planet.setBaseDefense(limit);
        }
    }
}

/** Edit planet Id. */
void
WSimPlanetEditor::editId()
{
    UIDecimalSelector select(0, 1, NUM_PLANETS, 10);
    select.setValue(planet.getId());
    if (select.doStandardDialog(_("Set Id"), _("Id:"), 0)) {
        planet.setId(select.getValue());
        setPlanetName(planet);
    }
}

/** Edit planet friendly code. */
void
WSimPlanetEditor::editFCode()
{
    /* build list of available codes */
    GFCodeList filtered_list;
    GFCodeList& full_list = GFCodeList::getInstance();

    for (GFCodeList::iterator i = full_list.begin(); i != full_list.end(); ++i) {
        if ((i->getFlags() & GFCode::fc_Planet) == 0)
            continue;
        filtered_list.addFCode(*i);
    }

    WFCodeWindow window(_("Set Friendly Code"), filtered_list);
    window.setFCode(planet.getFCode());

    ui_root->centerWidget(window);
    if (ui_root->runModal(window) == cm_Enter)
        planet.setFCode(window.getFCode());
}

/** Edit planetary defense. */
void
WSimPlanetEditor::editDefense()
{
    UIDecimalSelector select(0, 0, 600 /*FIXME: magic*/, 10);
    select.setValue(planet.getDefense());
    if (select.doStandardDialog(_("Set Defense"), _("Defense:"), 0))
        planet.setDefense(select.getValue());
}

/** Edit planetary population. */
void
WSimPlanetEditor::editPopulation()
{
    /* predictions for instruction text */
    int defense = planet.getDefense();
    if (defense < 10)
        defense = 60;
    int clans;
    if (defense < 50)
        clans = defense;
    else
        clans = 50 + square(defense - 50);

    /* build a window */
    ObjectHolder h;

    UIWindow& window = h.add(new UIWindow(_("Edit Population"), UIWindow::BLUE_WINDOW, COLOR_GRAY));
    window.setLayout(UIVBoxLayout::instance5);

    UIInputLine& edit = h.add(new UIInputLine(6, 10, 0, UIInputLine::ilf_NumbersOnly | UIInputLine::ilf_Framed, 0));

    window.add(h.add(new UIMultilineStatic
                     (FONT_NORMAL,
                      format(_("Enter population of the planet. PCC2 will then "
                               "compute the maximum possible number of defense "
                               "posts supported by that population. For example, "
                               "%d clans support up to %d defense posts."),
                             clans, defense),
                      ui_screen->getSize().x / 2)));

    UIGroup& group_edit = h.add(new UIGroup(UIHBoxLayout::instance5));
    group_edit.add(edit);
    group_edit.add(h.add(new UIStaticText(_("clans"), UIStaticText::StaticTitle)));
    window.add(group_edit);

    UIGroup& group_buttons = h.add(new UIGroup(UIHBoxLayout::instance5));
    group_buttons.add(h.add(new UIButton(cm_Enter, _("OK"), SDLK_RETURN)).addStop());
    group_buttons.add(h.add(new UIButton(cm_Escape, _("Cancel"), SDLK_ESCAPE)).addStop());
    group_buttons.add(h.add(new UISpacer()));
    window.add(group_buttons);

    window.pack();
    ui_root->centerWidget(window);
    int rv = ui_root->runModal(window);
    if (rv == cm_Enter) {
        int new_clans;
        if (stringValue(edit.getText(), new_clans)) {
            // FIXME: this duplicates the formula from
            // getMaxStructuresFormula, but this avoids >>100k
            // dependency in PlayVCR
            int new_defense;
            if (new_clans <= 50)
                new_defense = new_clans;
            else
                new_defense = 50 + int(std::sqrt(double(new_clans - 50)) + 0.5);
            planet.setDefense(new_defense);
        }
    }
}

/** Edit starbase beam tech. */
void
WSimPlanetEditor::editBaseBeamTech()
{
    UIStringList list;
    list.add(0, _("0 - No base"));
    for (int i = 1; i <= NUM_BEAMS; ++i)
        list.add(i, format(_("%X - %s"), i, getBeam(i).getName()));

    UIStandardListbox box(0, list.getCount(), getTextWidth(*fonts[FONT_NORMAL], list) + 20, list);
    box.scrollToKey(planet.getBaseBeamTech());
    if (box.doStandardListbox(_("Set Starbase Beam Tech"), 0))
        planet.setBaseBeamTech(box.getKey());
}

/** Edit number of starbase fighters. */
void
WSimPlanetEditor::editBaseFighters()
{
    if (planet.getBaseBeamTech() != 0) {
        UIDecimalSelector select(0, 0, config.MaximumFightersOnBase(planet.getOwner()), 10);
        select.setValue(planet.getBaseFighters());
        if (select.doStandardDialog(_("Set Base Fighters"), _("Fighters:"), 0))
            planet.setBaseFighters(select.getValue());
    }
}

/** Edit number of starbase defense. */
void
WSimPlanetEditor::editBaseDefense()
{
    if (planet.getBaseBeamTech() != 0) {
        UIDecimalSelector select(0, 0, config.MaximumDefenseOnBase(planet.getOwner()), 10);
        select.setValue(planet.getBaseDefense());
        if (select.doStandardDialog(_("Set Base Defense"), _("Defense:"), 0))
            planet.setBaseDefense(select.getValue());
    }
}

/** Edit starbase torpedo tech. */
void
WSimPlanetEditor::editBaseTorpTech()
{
    if (planet.getBaseBeamTech() != 0) {
        UIStringList list;
        for (int i = 1; i <= NUM_BEAMS; ++i)
            list.add(i, format(_("%X - %s"), i, getTorp(i).getName()));

        UIStandardListbox box(0, list.getCount(), getTextWidth(*fonts[FONT_NORMAL], list) + 20, list);
        box.scrollToKey(planet.getBaseTorpTech());
        if (box.doStandardListbox(_("Set Starbase Torp Tech"), 0))
            planet.setBaseTorpTech(box.getKey());
    }
}

class WSimPlanetTorpLine : public UIDecimalSelector {
    const int torpnr;
 public:
    WSimPlanetTorpLine(int torpnr, int value);
    ~WSimPlanetTorpLine();
    void drawContent(GfxCanvas& can);
    void getLayoutInfo(LayoutInfo& info);
};

WSimPlanetTorpLine::WSimPlanetTorpLine(int torpnr, int value)
    : UIDecimalSelector(torpnr, 0, 10000, 10),
      torpnr(torpnr)
{
    setValue(value);
}

WSimPlanetTorpLine::~WSimPlanetTorpLine()
{ }

void
WSimPlanetTorpLine::drawContent(GfxCanvas& can)
{
    string_t text_l = " " + getTorp(torpnr).getName();
    string_t text_r = numToString(getValue()) + " ";
    const UIColorSet& scheme = (getFocusState() != NoFocus ? BLACK_COLOR_SET : GRAY_COLOR_SET);

    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(GfxStandardColorScheme::instance);
    ctx.useFont(*fonts[FONT_NORMAL]);

    // Background
    drawSolidBar(ctx, extent, scheme[UIColor::tc_Background]);

    // length
    int len   = fonts[FONT_NORMAL]->getTextWidth(text_r);
    int split = extent.w - len;

    // draw right side first
    ctx.setColor(scheme[UIColor::tc_Static]);
    outTextF(ctx, extent.x + split, extent.y, len, text_r);
    outTextF(ctx, extent.x, extent.y, split, text_l);

    can.update(extent);
}

void
WSimPlanetTorpLine::getLayoutInfo(LayoutInfo& info)
{
    info.min_size = info.pref_size =
        GfxPoint(fonts[FONT_NORMAL]->getTextWidth(getTorp(torpnr).getName() + "  99999"),
                 font_heights[FONT_NORMAL]);
    info.flags = li_FlexibleH;
}

/** Edit starbase torps. */
void
WSimPlanetEditor::editBaseTorps()
{
    if (planet.getBaseBeamTech() != 0) {
        WSimPlanetTorpLine* torp_edits[NUM_TORPS];
        ObjectHolder h;
        UIWindow& w = h.add(new UIWindow(_("Starbase Torpedoes"), UIWindow::BLUE_WINDOW, COLOR_GRAY));
        w.setLayout(UIVBoxLayout::instance5);

        UIGroup& group_torps   = h.add(new UIGroup(UIVBoxLayout::instance0));
        UIGroup& group_buttons = h.add(new UIGroup(UIHBoxLayout::instance5));
        for (int i = 1; i <= NUM_TORPS; ++i) {
            WSimPlanetTorpLine& tl = h.add(new WSimPlanetTorpLine(i, planet.getBaseTorps(i)));
            torp_edits[i-1] = &tl;
            group_torps.add(tl);
        }
        group_torps.add(h.add(new UIFocusIterator(1, NUM_TORPS,
                                                  UIFocusIterator::fi_Vert | UIFocusIterator::fi_Page | UIFocusIterator::fi_Home)));
        group_buttons.add(h.add(new UIButton(cm_Enter, _("OK"), SDLK_RETURN)).addStop());
        group_buttons.add(h.add(new UIButton(cm_Escape, _("Cancel"), SDLK_ESCAPE)).addStop());
        group_buttons.add(h.add(new UISpacer()));
        w.add(group_torps);
        w.add(group_buttons);
        int focus_id = (planet.getBaseTorpTech() == 0 ? 1 : planet.getBaseTorpTech());
        torp_edits[focus_id-1]->setFocusedWidget();

        w.pack();
        ui_root->moveWidgetToEdge(w, 2, 2, 10);
        int rv = ui_root->runModal(w);
        if (rv == cm_Enter) {
            for (int i = 1; i <= NUM_TORPS; ++i)
                planet.setBaseTorps(i, torp_edits[i-1]->getValue());
        }
    }
}

/**************************** WSimOptionsTile ****************************/

// FIXME: this class is ugly.
class WSimOptionTile : public WTile {
    GSimOptions& opts;
    ObjectHolder h;

    static const char*const left_column[];
    static const char*const right_column[];
    static const char keys[];

    int left_x, left_wi;
    int right_x;

    void computeLayout();
 public:
    WSimOptionTile(GSimOptions& opts);

    bool handleEvent(const UIEvent& e, bool second_pass);
    void getLayoutInfo(LayoutInfo& info);
    void drawData(GfxCanvas& can);
};

/* Strings are enumerated here to simplify layout computation */
const char*const WSimOptionTile::left_column[] = {
    N_("Engine-Shield Bonus:"),
    N_("Fed Crew Bonus:"),
    N_("Random Left/Right:"),
    N_("Left/Right Balance:"),
    N_("VCR Algorithm:"),
    N_("Honor alliances/teams:"),
    N_("Battles per simulation:"),
    N_("Seed control:"),
    N_("Randomize FCodes:")
};

const char WSimOptionTile::keys[] = { 'E', 'C', 'R', 'L', 'P', 'A', 'B', 'S', 'F' };

/* Note: for simplicity, we assume that the first entries
   match VcrMode */
const char*const WSimOptionTile::right_column[] = {
    /* Host mode */
    N_("Host"),                 // 0
    N_("PHost 2"),
    N_("PHost 3"),
    N_("PHost 4"),
    N_("FLAK"),
    N_("NuHost"),               // 5
    0,
    0,
    0,
    0,

    /* ES-Bonus */
    N_("none"),                 // 10
    /* Boolean options */
    N_("Yes"),                  // 11
    N_("No"),                   // 12
    /* Balancing */
    N_("360 kt (Host)"),        // 13
    N_("Master at Arms"),       // 14
    /* battles */
    N_("one"),                  // 15
    N_("complete"),             // 16
    /* Random fcode */
    N_("once"),                 // 17
    N_("every time")            // 18
};

WSimOptionTile::WSimOptionTile(GSimOptions& opts)
    : WTile(GfxPoint(1, 1)), opts(opts)
{
    using std::tolower;
    const int ly = font_heights[FONT_TITLE];
    setLayout(LeftJustified, 0, false);
    for (uint_t i = 0; i < countof(keys); ++i) {
        addControl(GfxRect(0, ly*i+1, ly-2, ly-2),
                   h.add(new UIButton(string_t(&keys[i], 1),
                                      tolower(keys[i]))));
    }
}

bool
WSimOptionTile::handleEvent(const UIEvent& e, bool second_pass)
{
    if (e.type == ev_Key) {
        switch (e.key.code) {
         case 'e':
            // ESB
            {
                UIDecimalSelector select(0, 0, 500, 10);
                select.setValue(opts.engine_shield_bonus);
                if (select.doStandardDialog(_("Engine-Shield Bonus"), _("Rate [%]"), 0)) {
                    opts.engine_shield_bonus = select.getValue();
                    drawWidget();
                }
            }
            return true;
         case 'c':
            // Fed bonus
            opts.scotty_bonus = !opts.scotty_bonus;
            drawWidget();
            return true;
         case 'r':
            // random l/r
            opts.random_left_right = !opts.random_left_right;
            drawWidget();
            return true;
         case 'l':
            // balancing
            if (int(opts.balancing_mode) >= opts.BalanceMAX)
                opts.balancing_mode = GSimOptions::BalancingMode(0);
            else
                opts.balancing_mode = GSimOptions::BalancingMode(opts.balancing_mode + 1);
            drawWidget();
            return true;
         case 'p':
            // host
            while (1) {
                if (int(opts.vcr_mode) >= opts.VcrMAX)
                    opts.vcr_mode = GSimOptions::VcrMode(0);
                else
                    opts.vcr_mode = GSimOptions::VcrMode(opts.vcr_mode + 1);

                // do we accept this mode?
#ifndef CONF_FLAK_SUPPORT
                if (opts.vcr_mode == GSimOptions::VcrFLAK)
                    continue;
#endif
                break;
            }
            drawWidget();
            return true;
         case 'a':
            // allies
            opts.honor_alliances = !opts.honor_alliances;
            drawWidget();
            return true;
         case 'b':
            // battles per sim
            opts.only_one_simulation = !opts.only_one_simulation;
            if (!opts.only_one_simulation)
                opts.seed_control = false;
            drawWidget();
            return true;
         case 's':
            // seed control
            opts.seed_control = !opts.seed_control;
            if (opts.seed_control)
                opts.only_one_simulation = true;
            drawWidget();
            return true;
         case 'f':
            // random fcodes
            opts.randomize_fcodes_on_every_fight = !opts.randomize_fcodes_on_every_fight;
            drawWidget();
            return true;
        }
    }
    return WTile::handleEvent(e, second_pass);
}

void
WSimOptionTile::getLayoutInfo(LayoutInfo& info)
{
    computeLayout();
    info.min_size = info.pref_size =
        GfxPoint(right_x + 5, countof(left_column) * font_heights[FONT_TITLE]);
    info.flags = 0;
}

void
WSimOptionTile::drawData(GfxCanvas& can)
{
    // draw left side
    computeLayout();

    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(getSkin());
    ctx.useFont(*fonts[FONT_NORMAL]);
    ctx.setTextAlign(0, 1);
    ctx.setColor(UIColor::tc_Static);
    for (unsigned int i = 0; i < countof(left_column); ++i) {
        int y = (font_heights[FONT_TITLE] * (2*i+1)) / 2;
        outText(ctx, left_x, y, _(left_column[i]));
    }

    // draw right side
    int right_wi = right_x - left_x - left_wi;
    ctx.setTextAlign(2, 1);

    const int ly = font_heights[FONT_TITLE];
    int y = ly/2;

    ctx.setColor(UIColor::tc_Green);
    ctx.setSolidBackground();

    // ESB
    if (opts.engine_shield_bonus != 0)
        outTextF(ctx, right_x, y, right_wi, format("%d%%", opts.engine_shield_bonus));
    else
        outTextF(ctx, right_x, y, right_wi, _(right_column[10]));
    y += ly;

    // Crew Bonus
    outTextF(ctx, right_x, y, right_wi, _(right_column[opts.scotty_bonus ? 11 : 12]));
    y += ly;

    // Random left/right
    outTextF(ctx, right_x, y, right_wi, _(right_column[opts.random_left_right ? 11 : 12]));
    y += ly;

    // Left/right balance
    if (opts.balancing_mode == GSimOptions::Balance360k)
        outTextF(ctx, right_x, y, right_wi, _(right_column[13]));
    else if (opts.balancing_mode == GSimOptions::BalanceMasterAtArms)
        outTextF(ctx, right_x, y, right_wi, _(right_column[14]));
    else
        outTextF(ctx, right_x, y, right_wi, _(right_column[10]));
    y += ly;

    // VCR Algorithm
    outTextF(ctx, right_x, y, right_wi, _(right_column[opts.vcr_mode]));
    y += ly;

    // Honor alliances
    outTextF(ctx, right_x, y, right_wi, _(right_column[opts.honor_alliances ? 11 : 12]));
    y += ly;

    // Battles per simulation
    outTextF(ctx, right_x, y, right_wi, _(right_column[opts.only_one_simulation ? 15 : 16]));
    y += ly;

    // Seed control
    outTextF(ctx, right_x, y, right_wi, _(right_column[opts.seed_control ? 11 : 12]));
    y += ly;

    // Random fcodes
    outTextF(ctx, right_x, y, right_wi, _(right_column[opts.randomize_fcodes_on_every_fight ? 18 : 17]));
    y += ly;
}

void
WSimOptionTile::computeLayout()
{
    int right_wi;

    left_wi = getTextWidth(*fonts[FONT_NORMAL], StaticTranslatedStringTable(left_column, countof(left_column))) + 10;
    right_wi = getTextWidth(*fonts[FONT_NORMAL], StaticTranslatedStringTable(right_column, countof(right_column)));

    left_x = font_heights[FONT_TITLE] + 5;
    right_x = left_x + left_wi + right_wi + 5;
}

/******************************* WSimScreen ******************************/

GSimState WSimScreen::sim_state;
GSimOptions WSimScreen::sim_opts;

WSimScreen::WSimScreen()
    : UIWindow(_("Battle Simulator"), BLUE_WINDOW, COLOR_DARK),
      ed_list(h.add(new WSimListWithHandler(sim_state)))
{
    init();
}

WSimScreen::~WSimScreen()
{ }

void
WSimScreen::init()
{
    setLayout(UIVBoxLayout::instance5);

    UIGroup& all_main             = h.add(new UIGroup(UIHBoxLayout::instance5));
    UIGroup& all_buttons_1        = h.add(new UIGroup(UIHBoxLayout::instance5));
    UIGroup& all_buttons_2        = h.add(new UIGroup(UIHBoxLayout::instance5));
    UIGroup& all_main_list        = h.add(new UIGroup(UIVBoxLayout::instance5));
    UIGroup& all_main_list_button = h.add(new UIGroup(UIHBoxLayout::instance0));
    UIFrameGroup& all_panels      = h.add(new UIFrameGroup(UIVBoxLayout::instance5, 5));
    UIGroup& all_panels_group     = h.add(new UIGroup(UIVBoxLayout::instance5));
    UIGroup& ship_info_group      = h.add(new UIGroup(UIHBoxLayout::instance5));

    WSimPlanetButton& ed_planet = h.add(new WSimPlanetButton(sim_state));
    UIFrameGroup& ed_list_frame = h.add(new UIFrameGroup(UIVBoxLayout::instance0));
    ed_list_frame.add(ed_list);

    WSimShipWeaponTile& ssw_tile = h.add(new WSimShipWeaponTile(ed_list));
    ssw_tile.setHeading(_("Weapons"));
    ssw_tile.setDisplayState(WTile::sNormal);

    WSimShipTypeTile&   sst_tile = h.add(new WSimShipTypeTile(ed_list));
    WSimShipDamageTile& ssd_tile = h.add(new WSimShipDamageTile(ed_list));
    WSimShipImageTile&  ssi_tile = h.add(new WSimShipImageTile(ed_list));

    all_main_list_button.add(ed_planet);
    all_main_list_button.add(h.add(new UIButton("P", 'p')));

    all_main_list.add(h.add(new UIStaticText(_("Ships in simulation:"), UIStaticText::WhiteTitle)));
    all_main_list.add(ed_list_frame);
    all_main_list.add(all_main_list_button);

    ship_info_group.add(ssi_tile);
    ship_info_group.add(sst_tile);

    all_panels.add(ship_info_group);
    all_panels.add(ssd_tile);
    all_panels.add(ssw_tile);
    all_panels.add(h.add(new UISpacer()));
    if (sim_game_if->isGameAvailable()) {
        UIGroup& game_group = h.add(new UIGroup(UIHBoxLayout::instance5));
        game_group.add(h.add(new WSimShipButton(ed_list, "U", 'u')));
        game_group.add(h.add(new WSimShipButton(ed_list, "W", 'w')));
        game_group.add(h.add(new UIStaticText(_("Update/Write back"), UIStaticText::WhiteTitle)));
        game_group.add(h.add(new UISpacer()));
        game_group.add(h.add(new WSimShipButton(ed_list, "F1", SDLK_F1)));
        all_panels.add(game_group);
    }

    all_panels_group.add(h.add(new UIStaticText(_("This ship:"), UIStaticText::WhiteTitle)));
    all_panels_group.add(all_panels);

    all_main.add(all_main_list);
    all_main.add(all_panels_group);

    all_buttons_1.add(h.add(new UIButton(_("Ins - Add"), SDLK_INSERT)));
    all_buttons_1.add(h.add(new WSimShipButton(ed_list, _("Del - Remove"), SDLK_DELETE)));
    all_buttons_1.add(h.add(new WSimShipButton(ed_list, _("Ctrl-S - Save"), ss_Ctrl + 's')));
    all_buttons_1.add(h.add(new UISpacer()));
    all_buttons_1.add(h.add(new UIButton(_("Ctrl-O - Options"), ss_Ctrl + 'o')));

    all_buttons_2.add(h.add(new WSimShipButton(ed_list, _("ENTER - Simulate"), SDLK_RETURN)));
    all_buttons_2.add(h.add(new UIButton(0, _("ESC"), SDLK_ESCAPE)).addStop());
    all_buttons_2.add(h.add(new UIButton(_("Ctrl-R - Load"), ss_Ctrl + 'r')));
    // all_buttons_2.add(h.add(new UIButton(_("Ctrl-A - Allies"), ss_Ctrl + 'a')));
    all_buttons_2.add(h.add(new UISpacer()));
    all_buttons_2.add(h.add(new UIButton(_("Help"), 'h')));

    add(all_main);
    add(all_buttons_1);
    add(all_buttons_2);
    add(h.add(new WHelpWidget("pcc2:bsim")));

    pack();

    ed_list.sig_menu.add(this, &WSimScreen::onContextMenu);
}

/** Planet Editor. */
void
WSimScreen::editPlanet()
{
    GSimPlanet edit;
    if (sim_state.hasPlanet())
        edit = sim_state.getPlanet();
    setPlanetName(edit);

    WSimPlanetEditor editor(edit);
    ui_root->centerWidget(editor);
    int rv = ui_root->runModal(editor);

    switch (rv) {
     case cm_Enter:
        sim_state.addPlanet();
        sim_state.getPlanet() = edit;
        break;
     case cm_No:
        sim_state.removePlanet();
        break;
    }
}

/** Option editor. */
void
WSimScreen::editOptions()
{
    GSimOptions opts = sim_opts;

    ObjectHolder h;
    WSimOptionTile tile_opts(opts);

    UIWindow win(_("Simulator Options"), UIWindow::BLUE_WINDOW, COLOR_GRAY);
    win.setLayout(UIVBoxLayout::instance5);
    win.add(tile_opts);

    UIGroup group_buttons(UIHBoxLayout::instance5);
    UIButton button_ok(cm_Enter, _("OK"), SDLK_RETURN);
    UIButton button_cancel(cm_Enter, _("Cancel"), SDLK_ESCAPE);
    UISpacer spacer;
    UIButton button_help(_("Help"), 'h');
    WHelpWidget help("pcc2:simopts");
    button_ok.addStop();
    button_cancel.addStop();
    group_buttons.add(button_ok);
    group_buttons.add(button_cancel);
    group_buttons.add(spacer);
    group_buttons.add(button_help);
    win.add(group_buttons);
    win.add(help);
    win.pack();

    ui_root->centerWidget(win);
    int rv = ui_root->runModal(win);

    if (rv == cm_Enter)
        sim_opts = opts;
}

/** Load file. Ask for file name, then load the file and replace
    current selection with it. */
void
WSimScreen::loadFile()
{
    UIFileChooserDialog select(_("Load Simulation"),
                               Directory::openDir(Directory::getWorkingDirectory()),
                               FileNamePattern::getAllFilesWithExtensionPattern("ccb"));
    ui_root->centerWidget(select);
    bool rv = ui_root->runModal(select);

    if (rv) {
        string_t result = appendFileNameExtension(select.getResultFile(), "ccb", false);
        try {
            {
                Ptr<Stream> file = Stream::openFile(result, Stream::C_READ);
                GSimState tmp;
                tmp.load(*file);
                sim_state.swap(tmp);
            }
            if (sim_state.hasPlanet())
                setPlanetName(sim_state.getPlanet());
            sim_state.doUpdates();
            if (!sim_state.isMatchingShipList())
                messageBox(_("This simulation seems to have been set up with a different ship list. "
                             "To regenerate the intended results, restart the Simulator from the correct "
                             "game directory."),
                           _("Load Simulation"));
        }
        catch (FileFormatException& e) {
            // FIXME: could refer to e.getFileName() here
            messageBox(format(_("Unable to load %s: %s"), result, e.what()),
                       _("Load Simulation"));
        }
        catch (std::runtime_error& e) {
            messageBox(format(_("Unable to open %s: %s"), result, e.what()),
                       _("Load Simulation"));
        }
    }
}

/** Save file. */
void
WSimScreen::saveFile()
{
    UIFileChooserDialog select(_("Save Simulation"),
                               Directory::openDir(Directory::getWorkingDirectory()),
                               FileNamePattern::getAllFilesWithExtensionPattern("ccb"));
    ui_root->centerWidget(select);
    bool rv = ui_root->runModal(select);

    if (rv) {
        string_t result = appendFileNameExtension(select.getResultFile(), "ccb", false);
        try {
            Ptr<Stream> file = Stream::openFile(result, Stream::C_CREATE | Stream::C_WRITE);
            sim_state.save(*file);
        }
        catch (std::runtime_error& e) {
            messageBox(format(_("Unable to save %s: %s"), result, e.what()),
                       _("Save Simulation"));
        }
    }
}

/** Fleet cost comparison dialog. */
void
WSimScreen::doFleetCostComparison()
{
    if (sim_state.getNumShips() > 0) {
        WFleetCostDialog dlg(sim_state);
        ui_root->centerWidget(dlg);
        ui_root->runModal(dlg);
    }
}

/** Run the simulation (ENTER). Does the first simulation and opens
    the result window which allows users to run more simulations. */
void
WSimScreen::runSimulation()
{
    /* do we allow to run the simulation? */
    if (sim_state.getNumShips() == 0)
        return;
    if (sim_state.getNumShips() == 1 && !sim_state.hasPlanet())
        return;

    /* yes, we do. First randomize fcodes as requested. */
    if (!sim_opts.randomize_fcodes_on_every_fight) {
        sim_state.assignRandomFCodes();
        sim_state.doUpdates();
    }

    /* now try to get some results */
    WSimResultWindow result_window(sim_state, sim_opts);
    if (!result_window.runFirstSimulation()) {
        /* no results, figure out why and give a hopefully helpful hint. */
        bool have_disabled = false;
        for (GSimState::ship_index_t i = 0; i < sim_state.getNumShips(); ++i) {
            if ((sim_state.getShip(i).getFlags() & GSimObject::fl_Deactivated) != 0) {
                have_disabled = true;
                break;
            }
        }
        if (have_disabled) {
            messageBox(_("There are no fights in this arrangement. You have disabled "
                         "some ships, try re-enabling them using [.]."),
                       _("Simulator"));
        } else {
            messageBox(_("There are no fights in this arrangement. Did you set all "
                         "owners and aggression settings correctly?"),
                       _("Simulator"));
        }
    } else {
        /* okay, let's boogie */
        ui_root->centerWidget(result_window);
        int value = ui_root->runModal(result_window);
        if (value == cm_Enter) {
            int index = result_window.getResultListIndex();
            if (index >= sim_state.getNumShips() && sim_state.hasPlanet())
                editPlanet();
            else
                ed_list.scrollTo(index, 1);
        }
    }
}

/** Write all ships to game (Ctrl-W). */
void
WSimScreen::writeToGame()
{
    /* only try when a game is available */
    if (!sim_game_if->isGameAvailable())
        return;

    int count = 0, failed = 0;
    for (GSimState::ship_index_t i = 0; i < sim_state.getNumShips(); ++i) {
        const GSimShip& sh = sim_state.getShip(i);
        if (sim_game_if->getShipRelation(sh) == GSimulatorGameInterface::ShipIsPlayed) {
            if (sim_game_if->updateToGame(sh))
                ++count;
            else
                ++failed;
        }
    }

    if (count > 0)
        messageBox(format(_("%d ship%!1{s have%| has%} been written back to the game."), count),
                   _("Simulator"));
    else if (failed > 0)
        messageBox(format(_("%d ship%!1{s%} could not be written back to the game."), failed),
                   _("Simulator"));
    else
        messageBox(_("There are none of your ships in this simulation."),
                   _("Simulator"));
}

/** Update all ships from game (Ctrl-U). */
void
WSimScreen::updateFromGame()
{
    /* only try when a game is available */
    if (!sim_game_if->isGameAvailable())
        return;

    int count = 0, failed = 0;
    for (GSimState::ship_index_t i = 0; i < sim_state.getNumShips(); ++i) {
        GSimShip& sh = sim_state.getShip(i);
        if (sim_game_if->getShipRelation(sh) != GSimulatorGameInterface::ShipIsUnknown) {
            if (sim_game_if->updateFromGame(sh))
                ++count;
            else
                ++failed;
        }
    }

    if (count > 0)
        messageBox(format(_("%d ship%!1{s have%| has%} been updated from the game."), count),
                   _("Simulator"));
    else if (failed > 0)
        messageBox(format(_("%d ship%!1{s%} could not be updated from the game."), failed),
                   _("Simulator"));
    else
        messageBox(_("There are none of your ships in this simulation."),
                   _("Simulator"));
}

bool
WSimScreen::handleEvent(const UIEvent& event, bool second_pass)
{
    if (event.type == ev_Key) {
        switch (event.key.code) {
         case 'p':
            editPlanet();
            sim_state.doUpdates();
            return true;
         case ss_Ctrl + 'w':
            writeToGame();
            return true;
         case ss_Ctrl + 'u':
            updateFromGame();
            sim_state.doUpdates();
            return true;
         case ss_Ctrl + 'o':
            editOptions();
            return true;
         case ss_Ctrl + 'r':
            loadFile();
            return true;
         case ss_Ctrl + 's':
            saveFile();
            return true;
         case ss_Ctrl + 'c':
            doFleetCostComparison();
            return true;
         case ss_Ctrl + SDLK_DELETE:
            if (UIMessageBox(_("Clear this simulation arrangement?"), _("Battle Simulator")).doYesNoDialog()) {
                GSimState().swap(sim_state);
                sim_state.doUpdates();
            }
            return true;
         case '\\':
         case '#':
            onContextMenu();
            return true;
         case SDLK_RETURN:
            runSimulation();
            return true;
         // Missing from PCC 1.x:
         //   F4, k (FLAK rating override), help
         // Missing from CCBSim 1.x:
         //   C-v (VCR view), C-l (load pconfig), M-i (version)
        }
    }
    return UIWindow::handleEvent(event, second_pass);
}

void
WSimScreen::onContextMenu()
{
    /* No context menu if we don't have anything on our list */
    if (sim_state.getNumShips() == 0)
        return;

    StringList sl;
    sl.add(0, _("Sort by Id"));
    sl.add(1, _("Sort by Owner"));
    sl.add(2, _("Sort by Hull"));
    sl.add(3, _("Sort by Battle Order"));
    UIStandardListbox menu(0, 1, getTextWidth(*fonts[FONT_NORMAL], sl) + 20, sl);
    if (menu.doStandardMenu(ed_list)) {
        switch (menu.getKey()) {
         case 0:
            sortBy(sortById);
            break;
         case 1:
            sortBy(sortByOwner);
            break;
         case 2:
            sortBy(sortByHull);
            break;
         case 3:
            sortBy(sortByBattleOrder);
            break;
        }
        sim_state.doUpdates();
    }
}

void
WSimScreen::sortBy(bool compare(const GSimShip& a, const GSimShip& b))
{
    /* Anything to do? */
    if (sim_state.getNumShips() == 0)
        return;

    /* Remember old selection */
    int oldId = sim_state.getShip(ed_list.getCurrent()).getId();
    
    /* Bubblesort. Who cares. */
    GSimState::ship_index_t n = 1;
    while (n < sim_state.getNumShips()) {
        if (compare(sim_state.getShip(n), sim_state.getShip(n-1))) {
            sim_state.swapShips(n, n-1);
            if (n > 1)
                --n;
            else
                ++n;
        } else {
            ++n;
        }
    }

    /* Scroll to new selection */
    ed_list.scrollTo(sim_state.getIndexOf(oldId), 1);
    sim_state.doUpdates();
}

void
initSimulatorState()
{
    // Discard old state
    doneSimulatorState();

    // Set up configuration
    WSimScreen::sim_opts.loadDefaults(!host.isPHost() ? GSimOptions::VcrHost :
                                      host.host_version < MKVERSION(3,0,0) ? GSimOptions::VcrPHost2 :
                                      host.host_version < MKVERSION(4,0,0) ? GSimOptions::VcrPHost3 : GSimOptions::VcrPHost4);
}

void
doneSimulatorState()
{
    WSimScreen::sim_opts = GSimOptions();
    GSimState().swap(WSimScreen::sim_state);
}

int
doSimulator(GSimulatorGameInterface& game_iface)
{
    sim_game_if = &game_iface;

    if (WSimScreen::sim_state.hasPlanet())
        setPlanetName(WSimScreen::sim_state.getPlanet());

    try {
        WSimScreen me;
        me.setExtent(ui_root->getExtent());
        int rv = ui_root->runModal(me);

        sim_game_if = 0;

        return rv;
    }
    catch (...) {
        sim_game_if = 0;
        throw;
    }
}

/***************************** Initialisation ****************************/

extern const ModuleInitializer mod_sim[] = {
    { OnEnteredGame, initSimulatorState },
    { OnLeftGame, doneSimulatorState },
    { 0, 0 },
};
