/**
  *  \file client/widgets/lrupopup.cc
  *  \brief LRU List Popup
  */

#include <algorithm>
#include "client/widgets/lrupopup.h"
#include "ui/listbox.h"
#include "ui/uiglobal.h"
#include "gfx/gfxcontext.h"

namespace {
    /** Item for WLRUListBox. */
    struct WLRUNode {
        string_t name;          ///< Displayed name of item.
        LRUList* list;          ///< List this item is in, null if this is a heading.
        uint32_t index;         ///< Index into this item's list, don't-care if this is a heading.
        WLRUNode(string_t name, LRUList* list, uint32_t index)
            : name(name), list(list), index(index)
            { }
    };
    /** List of WLRUNode. */
    class WLRUListBox : public UIListbox {
     public:
        WLRUListBox(std::vector<WLRUNode>& nodes);
        void drawPart(GfxCanvas& can, int from, int to);
        bool isAccessible(int item);
     private:
        std::vector<WLRUNode>& nodes;
    };
}

WLRUListBox::WLRUListBox(std::vector<WLRUNode>& nodes)
    : UIListbox(0, std::min(3, std::max(20, int(nodes.size()))), 30*font_ems[FONT_NORMAL], nodes.size()),
      nodes(nodes)
{ }

void
WLRUListBox::drawPart(GfxCanvas& can, int from, int to)
{
    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useFont(*fonts[FONT_NORMAL]);

    for (int i = from; i < to; ++i) {
        GfxRect area;
        prepareEntry(ctx, ColorStyle, i, &area);

        int item = getTop() + i;
        if (item < int(nodes.size())) {
            const WLRUNode& node = nodes[item];
            if (node.list == 0) {
                int y = area.y + font_heights[FONT_NORMAL]/2 - 1;
                ctx.setColor(UIColor::tc_Faded);
                drawHLine(ctx, area.x + 2, y, area.x + 28);
                drawHLine(ctx, area.x + 32 + fonts[FONT_NORMAL]->getTextWidth(node.name), y, area.x + area.w - 2);
                outTextF(ctx, area.x + 30, area.y, area.w - 30, node.name);
            } else {
                outTextF(ctx, area.x + 10, area.y, area.w - 10, node.name);
            }
        }
    }
}

bool
WLRUListBox::isAccessible(int item)
{
    return item < int(nodes.size())
        && nodes[item].list != 0;
}

/******************************* WLRUPopup *******************************/

/** Constructor. */
WLRUPopup::WLRUPopup()
{ }

/** Add a list.
    \param name Name (heading)
    \param list List
    \return *this, for chaining of calls */
WLRUPopup&
WLRUPopup::withList(string_t name, LRUList& list)
{
    names.push_back(name);
    lists.push_back(&list);
    return *this;
}

/** Display popup list.
    \param current Current value. Defines the initial cursor position
    \param anchor Anchor widget, i.e. invoking button
    \return Selected list item, null if canceled or none */
const LRUListItem*
WLRUPopup::doPopup(string_t current, UIWidget& anchor)
{
    // Figure out whether we have to use headings
    int numNonEmptyLists = 0;
    for (uint32_t i = 0; i < lists.size(); ++i) {
        if (!lists[i]->empty()) {
            ++numNonEmptyLists;
        }
    }
    if (numNonEmptyLists == 0) {
        // Nothing to show
        return 0;
    }

    // Build the list
    std::vector<WLRUNode> userList;
    uint32_t userPos = 0;
    for (uint32_t i = 0; i < lists.size(); ++i) {
        LRUList& list = *lists[i];
        if (!list.empty()) {
            if (numNonEmptyLists != 1) {
                userList.push_back(WLRUNode(names[i], 0, 0));
            }
            for (uint32_t j = 0; j < list.size(); ++j) {
                if (userPos == 0 && LRUList::isSameValue(current, list[j].getValue())) {
                    userPos = userList.size();
                }
                userList.push_back(WLRUNode(list[j].getName(), &list, j));
            }
        }
    }

    // Make the list box and operate it
    WLRUListBox box(userList);
    box.scrollTo(userPos, 1);
    if (box.doStandardMenu(anchor)) {
        const WLRUNode& node = userList[box.getCurrent()];
        if (node.list != 0) {
            // item selected
            return &(*node.list)[node.index];
        } else {
            // shouldn't happen
            return 0;
        }
    } else {
        // canceled
        return 0;
    }
}
