/**
  *  \file game/fcode-game.cc
  *  \brief Friendly Codes -- Game-related Routines
  *
  *  This contains routines that interface game objects, to avoid
  *  dependencies on the whole game machinery by modules needing only
  *  friendly code list functionality.
  *
  *  (c) 2001-2012 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "game/fcode.h"
#include "game/planet.h"
#include "game/ship.h"
#include "game/spec.h"
#include "game/config.h"

/** Check whether code works on specified object.
    \param o a ship or planet */
bool
GFCode::worksOn(const GObject& o) const
{
    if (const GShip* s = dynamic_cast<const GShip*>(&o))
        return worksOn(*s);
    else if (const GPlanet* p = dynamic_cast<const GPlanet*>(&o))
        return worksOn(*p);
    else
        return false;
}

/** Check whether code works on ship.
    \param s a ship */
bool
GFCode::worksOn(const GShip& s) const
{
    if (!races.contains(raceId(s.getOwner())))
        return false;
    if ((flags & fc_Ship) == 0)
        return false;
    if ((flags & fc_Capital) && (s.getBeamCount() == 0)
        && (s.getTorpLauncherCount() == 0) && (s.getBayCount() == 0))
        return false;
    if ((flags & fc_Alchemy) && !s.canDoSpecial(hf_MerlinAlchemy)
        && !s.canDoSpecial(hf_NeutronicRefinery)
        && !s.canDoSpecial(hf_AriesRefinery))
        return false;
    return true;
}

/** Check whether code works on planet.
    \param p a planet */
bool
GFCode::worksOn(const GPlanet& p) const
{
    if (!races.contains(raceId(p.getOwner())))
        return false;
    if ((flags & fc_Planet) || ((flags & fc_Base) && p.hasBase()))
        return true;
    return false;
}


/** Make sublist of some other list. The new list will contain all
    friendly codes valid for object o.
    \param l   output list, should be empty
    \param o   object to generate list for */
GFCodeList::GFCodeList(const GFCodeList& l, const GObject& o)
{
    for (iterator i = l.begin(); i != l.end(); ++i)
        if (i->worksOn(o))
            addFCode(*i);
    sort();
}
