/**
  *  \file game/objtypes.cc
  *  \brief Object types
  *
  *  (c) 2001-2012 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "game/objtypes.h"
#include "game/obj.h"
#include "cpluslib/assert.h"

/****************************** GObjectType ******************************/

/** \class GObjectType
    \brief Object Type

    A class derived from GObjectType defines a set of objects, and
    iteration through them. In addition, it contains a general
    interface for object selection. Control screens and various other
    features use these as a back-end.

    A GObjectType object also manages a "current position".

    A GObjectType object starts without being associated to a universe.
    It can be associated with one by calling setUniverse(). The only
    derived-class method that has to deal with a null universe is init().
    The other methods can assume that a universe is present. Note that
    this implies the user shouldn't call any interesting method while
    the GObjectType is not associated with a universe.

    \todo We currently do not interpret any meaning into object Ids other
    than that 0 means no object. Do we need to be more strict? I guess we
    don't, and can just let GPlanetType etc. guarantee the additional
    requirement that id == got.getObject(id).getId(). Not having this
    requirement simplifies implementing the list for the visual scanner,
    for example. */

/** Initialize.
    \param max  Maximum object Id.
    \param trn  Related universe object. */
GObjectType::GObjectType()
    : current(0), universe(0)
{ }

GObjectType::~GObjectType()
{ }

/** Get current object. Returns 0 if this type doesn't have a
    universe, or there is no current object. */
GObject*
GObjectType::getCurrentObject()
{
    if (universe != 0 && isExistant(current)) {
        return &getObject(current);
    } else {
        return 0;
    }
}

/** Get next object after index. Can filter marked objects.
    \param index Index to start search at.
    \param mark true to return only marked objects */
int
GObjectType::getNextNoWrap(int index, bool mark)
{
    do
        index = getNextIndex(index);
    while (index != 0 && (!isExistant(index) || (mark && !getObject(index).isMarked())));
    return index;
}

/** Get previous object before index. Can filter marked objects.
    \param index Index to start search at.
    \param mark true to return only marked objects */
int
GObjectType::getPreviousNoWrap(int index, bool mark)
{
    do
        index = getPreviousIndex(index);
    while (index != 0 && (!isExistant(index) || (mark && !getObject(index).isMarked())));
    return index;
}

/** Get next object after index, with wrap. If the last object of a kind is
    reached, search starts again at the beginning. Can filter marked objects.
    \param index Index to start search at.
    \param mark true to return only marked objects */
int
GObjectType::getNextWrap(int index, bool mark)
{
    int n = getNextNoWrap(index, mark);
    if (n == 0)
        n = getNextNoWrap(0, mark);
    return n;
}

/** Get previous object before index, with wrap. If the last object of a kind is
    reached, search starts again at the beginning. Can filter marked objects.
    \param index Index to start search at.
    \param mark true to return only marked objects */
int
GObjectType::getPreviousWrap(int index, bool mark)
{
    int n = getPreviousNoWrap(index, mark);
    if (n == 0)
        n = getPreviousNoWrap(0, mark);
    return n;
}

/** Check for unit type.
    \retval true if this object type contains exactly one unit */
bool
GObjectType::isUnit()
{
    return getCurrent() != 0
        && getNextWrap(getCurrent(), false) == getCurrent();
}

/** Count objects. */
int
GObjectType::getNumObjects()
{
    int n = 0;
    for (int i = getNextNoWrap(0, false); i != 0; i = getNextNoWrap(i, false))
        ++n;
    return n;
}

/** Find nearest object. */
int
GObjectType::findNearest(const GPoint pt)
{
    int found = 0;
    int32_t dist2 = 0;
    bool inside = false;

    for (int i = getNextNoWrap(0, false); i != 0; i = getNextNoWrap(i, false)) {
        GMapObject* mo = dynamic_cast<GMapObject*>(&getObject(i));
        if (mo != 0) {
            GCircularObject* co = dynamic_cast<GCircularObject*>(mo);

            long ndist2 = distanceSquared(mo->getPos(), pt);
            bool ninside = (co != 0 && (ndist2 <= co->getRadiusSquared()));

            /* Pick this object, if
               - it's the first one
               - it's closer than our previous choice
               - we're inside it but not inside the previous one */
            if (found <= 0
                || ndist2 < dist2
                || (ninside && !inside))
            {
                found  = i;
                dist2  = ndist2;
                inside = ninside;
            }
        }
    }
    return found;
}

/** Make this type empty. To be called by implementors when they
    detect their type to be empty. */
void
GObjectType::setEmpty()
{
    ASSERT(getNextNoWrap(0, false) == 0);
    if (current != 0) {
        current = 0;
        sig_pos_changed.raise(0);
    }
}

/** Set current object Id.
    \pre isExistant(index) */
void
GObjectType::setCurrent(int index)
{
    // FIXME: this is called with index=0 and the assumption that this is a no-op. Better fix that someday.
    if (index != 0) {
        ASSERT(isExistant(index));
        if (current != index) {
            current = index;
            sig_pos_changed.raise(index);
        }
    }
}

/** Change universe. This can also be used to trigger re-scan of an existing universe. */
void
GObjectType::setUniverse(GUniverse* univ)
{
    this->universe = univ;
    this->init();
    if (univ == 0) {
        // Universe removed
        current = 0;
    } else {
        // Universe changed
        if (!isExistant(current))
            current = getNextNoWrap(0, false);
    }
    sig_pos_changed.raise(current);
}

/*************************** GIndexedObjectType **************************/

/** \class GIndexedObjectType
    \brief Indexed Object Type

    This refines GObjectType to represent a set of objects represented as a
    (sparse) object array. Objects are arranged in Id order. There is a finite
    set of Id numbers ranging from 1 to a given maximum. */

/** Constructor.
    \param max Highest possible Id number. */
GIndexedObjectType::GIndexedObjectType(int max)
    : max(max)
{ }

void
GIndexedObjectType::init()
{ }

int
GIndexedObjectType::getNextIndex(int index)
{
    if (++index <= max) {
        return index;
    } else {
        return 0;
    }
}

int
GIndexedObjectType::getPreviousIndex(int index)
{
    if (index == 0) {
        return max;
    } else {
        return index-1;
    }
}

/**************************** GAliasObjectType ***************************/

/** \class GAliasObjectType
    \brief Alias Object Type

    An object type that behaves just like another one, but has its
    own selection cursor. */

GAliasObjectType::GAliasObjectType(GObjectType& other)
    : other(other)
{
    conn_pos = other.sig_pos_changed.add(this, &GAliasObjectType::onChange);
    setCurrent(other.getCurrent());
    onChange();
}

void
GAliasObjectType::onChange()
{
    /* We must keep a correct Universe object; without one, user code does not
       believe we contain correct objects. Hence we copy other's universe whenever
       something changes. setUniverse() will call init(), which is a no-op for
       us, and it will call our sig_pos_changed. */
    setUniverse(other.getUniverse());
}

void
GAliasObjectType::init()
{ }

bool
GAliasObjectType::isExistant(int index)
{
    return other.isExistant(index);
}

GObject&
GAliasObjectType::getObject(int index)
{
    return other.getObject(index);
}

int
GAliasObjectType::getNextIndex(int index)
{
    return other.getNextNoWrap(index, false);
}

int
GAliasObjectType::getPreviousIndex(int index)
{
    return other.getPreviousNoWrap(index, false);
}

/************************** GObjectTypeObserver **************************/

/** \class GObjectTypeObserver
    \brief Object Type Observer

    This class observes a GObjectType, providing an overridable
    callback function onObjectChanged() whenever the current object
    of that type changes, or a different object is selected. */
    

/** Constructor.
    \param type Object Type to observe. */
GObjectTypeObserver::GObjectTypeObserver(GObjectType& type)
    : type(type)
{
    conn_pos = type.sig_pos_changed.add(this, &GObjectTypeObserver::onPositionChanged);
    onPositionChanged();
}

GObjectTypeObserver::~GObjectTypeObserver()
{ }

/** Called when a different object is selected on the GObjectType. */
void
GObjectTypeObserver::onPositionChanged()
{
    // reconnect object change signal, and update
    if (GObject* o = getObject()) {
        conn_obj = o->sig_changed.add(this, &GObjectTypeObserver::onObjectChanged);
    } else {
        conn_obj = 0;
    }
    onObjectChanged();
}

/** Called when a different object is selected on the GObjectType, or the object changes.
    This is a virtual method intended to be overridden. It cannot be abstract because
    the first call is implicitly made during the constructor. In case you're keeping
    state, you should probably call it from your constructor as well. */
void
GObjectTypeObserver::onObjectChanged()
{ }

/** Get pointer to referred object. Note that this can be 0 when the GObjectType has
    no universe or the universe has no object of the required type. Ideally, this
    should be used as
    <code>if (GFoo* p = dynamic_cast<GFoo*>(getObject())) { ... }</code>;
    this way, it automatically copes with lacking objects or objects of different
    types. */
GObject*
GObjectTypeObserver::getObject()
{
    return type.getCurrentObject();
}
