/**
  *  \file int/if/userif.cc
  *  \brief Interpreter: User Interface Binding
  *
  *  Rule 1: All screens that can be invoked by the script interpreter
  *  must start with a call to runRunnableProcesses /
  *  killTerminatedProcesses. This is used to continue the process
  *  that invoked the screen.
  *
  *  Rule 2: All screens that invoke the script interpreter (this
  *  includes screens that follow rule 1) must check for a screen
  *  change request immediately after the invocation, and process that
  *  request.
  *
  *  IntUserInterfaceBinding just signal their request to their
  *  caller. The caller must check this by Rule 2. In the simplest
  *  form, they perform a simple UI "stop()". If a screen can signal a
  *  request to its parent, that parent must also check for
  *  appropriate signals. This applies, for example, to everyone who
  *  can display a console.
  *
  *  (c) 2001-2012 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <stdio.h>
#include "int/if/userif.h"
#include "int/if/ifutil.h"
#include "int/interror.h"

static IntUserInterfaceBinding* ui_binding;

IntUserInterfaceBinding::IntUserInterfaceBinding()
    : prev(ui_binding),
      self(&ui_binding)
{
    if (prev)
        prev->self = &prev;
    ui_binding = this;
}

IntUserInterfaceBinding::~IntUserInterfaceBinding()
{
    *self = prev;
    if (prev)
        prev->self = self;
}

IntUserInterfaceBinding*
IntUserInterfaceBinding::getPreviousBinding()
{
    return prev;
}

IntUserInterfaceBinding*
IntUserInterfaceBinding::getCurrentInstance()
{
    return ui_binding;
}

void
IntUserInterfaceBinding::setProperty(IntUserInterfaceProperty p, IntValue* v)
{
    int32_t value, n;
    switch (p) {
     case iuiScanX:
        if (checkIntArg(value, v, 0, 10000) && (n = getProperty(iuiScanY)) >= 0)
            doMoveScanner(GPoint(value, n));
        break;
     case iuiScanY:
        if (checkIntArg(value, v, 0, 10000) && (n = getProperty(iuiScanX)) >= 0)
            doMoveScanner(GPoint(n, value));
        break;
     default:
        throw IntError::notAssignable();
    }
}

/*************************** A global instance ***************************/

namespace {
    class IntGlobalUserInterfaceBinding : public IntUserInterfaceBinding {
        int  getProperty(IntUserInterfaceProperty /*p*/)
            { return -1; }
        IntKeymapRef getKeymap()
            { return 0; }
#if 0
        void doGoToScreen(int /*screen*/, int /*obj*/)
            { }
        void doGoToChart(GPoint /*pt*/)
            { }
        void doMoveScanner(GPoint /*pt*/)
            { }
        void doPopupConsole()
            { }
        void doOverlayMessage(string_t /*msg*/)
            { }
        void doLeaveRace()
            { }
        void doSetChartView(string_t /*name*/)
            { }
#else
        void doGoToScreen(int screen, int obj)
            { printf("UI::doGoToScreen(%d,%d)\n", screen, obj); }
        void doGoToChart(GPoint pt)
            { printf("UI::doGoToChart(%d,%d)\n", pt.x, pt.y); }
        void doMoveScanner(GPoint pt)
            { printf("UI::doMoveScanner(%d,%d)\n", pt.x, pt.y); }
        void doPopupConsole()
            { printf("UI::doPopupConsole()\n"); }
        void doOverlayMessage(string_t msg)
            { printf("UI::doOverlayMessage('%s')\n", msg.c_str()); }
        void doLeaveRace()
            { printf("UI::doLeaveRace()\n"); }
        void doSetChartView(string_t name)
            { printf("UI::doSetChartView('%s')\n", name.c_str()); }
#endif
    };
}

IntGlobalUserInterfaceBinding theGlobalUIBinding;
