/**
  *  \file int/mutex.cc
  *  \brief Interpreter: Mutexes (Locks)
  *
  *  Mutexes/locks are a convention to provide cooperation-based mutual
  *  exclusion for game features. It is intended to avoid that the user
  *  accidentally cancels an Auto Task action, or vice versa.
  *
  *  As of 1.1.18, PCC honors the following locks:
  *  - "Snnn.WAYPOINT" (ship #nnn's waypoint and intercept mission)
  *  - "Pnnn.TAX" (planet #nnn's taxation)
  *  - "Pnnn.BUILD" (planet #nnn's structure building)
  *
  *  The scripting language provides a feature
  *      With Lock("...") Do
  *  causing the content of the 'With' block to be executed with a lock
  *  being held. Only one process can be in such a block, another process
  *  (or the same process again) trying to enter the block will fail.
  *
  *  When this pattern is followed, lifetime of locks will be nicely
  *  managed by the interpreter. When a process goes away, all its
  *  data will go away, and so does the lock. However, it is possible
  *  to do things like
  *      sharedVar := Lock("...")
  *  which will associate the lock with this process, but then leave a
  *  reference in the shared variable. The lock will thus be still active
  *  but its process will be gone. This is called an orphaned / disowned
  *  lock.
  *
  *  Note that we use the terms 'Lock' and 'Mutex' interchangably when
  *  talking about these things. 'Lock' is the user-visible name. However,
  *  there is also the internal feature 'lock' meaning "locking a sensor
  *  beam onto an object".
  *
  *  This does not implement PCC 1.x's "CC$Lock" and "CC$Unlock" commands.
  *  When needed, they can be implemented as scripts using something like
  *  a hash-of-mutexes.
  *
  *  (c) 2001-2012 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "int/mutex.h"
#include "int/interror.h"
#include "util/nls.h"
#include "int/value.h"
#include "int/names.h"
#include "cpluslib/ptrvector.h"
#include "int/args.h"
#include "int/if/ifutil.h"
#include "int/compound.h"
#include "int/intglobal.h"
#include "int/contextio.h"
#include "int/vmio.h"
#include "io/io.h"
#include "int/exec.h"
#include "io/streams.h"

/** All mutex names. */
static IntVariableNames mutex_names;

/** All mutex objects. */
static ptr_vector<IntMutex> mutex_objects;

/******************************** IntMutex *******************************/

/** Constructor.
    \param slot Slot in mutex_objects.
    \param note User's note
    \param owner Owning process */
inline
IntMutex::IntMutex(const uint32_t slot, const string_t& note, IntExecutionContext*const owner)
    : reference_counter(1),
      slot(slot),
      note(note),
      owner(owner)
{ }

IntMutex::~IntMutex()
{ }

/** Create a mutex. This assumes that the mutex is actually free.
    \param slot Slot in mutex_objects.
    \param note User's note
    \param owner Owning process */
IntMutex*
IntMutex::createMutex(const uint32_t slot, const string_t& note, IntExecutionContext*const owner)
{
    // Make room
    while (mutex_objects.size() <= slot) {
        mutex_objects.push_back(0);
    }

    // Make a new mutex
    IntMutex* mtx = new IntMutex(slot, note, owner);
    mutex_objects.replace_elem(slot, mtx);
    return mtx;
}

/** Remove reference. Unregisters the mutex and deletes it when this
    was the last reference. */
void
IntMutex::removeReference()
{
    if (--reference_counter == 0) {
        // replace_elem will delete this object
        mutex_objects.replace_elem(slot, 0);
    }
}

/** Get mutex name. */
const string_t&
IntMutex::getName() const
{
    return mutex_names.lookupAddress(slot);
}

/** Create new lock.
    \param name Name of lock, in upper-case
    \param note User's comment
    \param owner Owning process
    \return newly-created IntMutex object
    \throw IntError if there is a locking conflict */
IntMutex*
IntMutex::create(const string_t& name, const string_t& note, IntExecutionContext*const owner)
{
    // Get slot for this mutex. An existing slot will be recycled.
    uint32_t slot = mutex_names.addMaybe(name);

    // Check existing mutex
    if (slot < mutex_objects.size() && mutex_objects[slot] != 0) {
        throw IntError("Already locked");
    }

    // Create new mutex. This will place it in mutex_objects.
    return createMutex(slot, note, owner);
}

/** Query lock.
    \param name Name of lock, in upper-case
    \return Pointer to existing lock of that name, 0 if none */
IntMutex*
IntMutex::query(const string_t& name)
{
    IntVariableNames::index_t slot = mutex_names.lookup(name);
    if (slot == mutex_names.nil) {
        return 0;
    } else {
        return mutex_objects[slot];
    }
}

/** Disown/orphan all locks owned by a process.
    \param process Process to check for */
void
IntMutex::disownLocksByProcess(IntExecutionContext*const owner)
{
    for (uint32_t i = 0, e = mutex_objects.size(); i < e; ++i) {
        if (mutex_objects[i] != 0 && mutex_objects[i]->getOwner() == owner) {
            mutex_objects[i]->setOwner(0);
        }
    }
}

/** Enumerate mutexes. This will enumerate all mutexes as defined by
    the owner filter, and add them to data. Note that this will not
    add references to the mutexes.

    \param data [out] List will be produced here
    \param owner [in] Filter. If non-null, only list mutexes owned by
    this process. If null, list all mutexes. */
void
IntMutex::enumMutexes(std::vector<IntMutex*>& data, IntExecutionContext* owner)
{
    for (uint32_t i = 0, e = mutex_objects.size(); i < e; ++i) {
        if (mutex_objects[i] != 0
            && (owner == 0 || mutex_objects[i]->getOwner() == owner))
        {
            data.push_back(mutex_objects[i]);
        }
    }
}

/**************************** IntMutexContext ****************************/

/** Constructor.
    \param mtx Mutex. Must have one reference allocated to this object. */
IntMutexContext::IntMutexContext(IntMutex* mtx)
    : mtx(mtx)
{ }

/** Destructor. Removes the mutex when this was the last reference. */
IntMutexContext::~IntMutexContext()
{
    mtx->removeReference();
}

/** Lookup implementation. Mutex has no properties. */
int32_t
IntMutexContext::lookup(const string_t& /*name*/)
{
    return -1;
}

/** Set implementation. Mutex has no properties. */
void
IntMutexContext::set(int32_t /*index*/, IntValue* /*value*/)
{ }

/** Get implementation. Mutex has no properties. */
IntValue*
IntMutexContext::get(int32_t /*index*/)
{
    return 0;
}

/** Next implementation. Mutex is not iterable. */
bool
IntMutexContext::next()
{
    return false;
}

/** Clone implementation. */
IntContext*
IntMutexContext::clone()
{
    return new IntMutexContext(mtx->addReference());
}

/** GetObject implementation. Mutex has no object. */
GObject*
IntMutexContext::getObject()
{
    return 0;
}

/** EnumProperties implementation. Mutex has no properties. */
void
IntMutexContext::enumProperties(IntPropertyAcceptor& /*acceptor*/)
{ }

/** toString implementation. */
string_t
IntMutexContext::toString(bool /*readable*/)
{
    // FIXME: we can do better for readable=true
    return "#<lock>";
}

/** Store implementation.
    \param sv [out] Tag node to store to
    \param aux [out] Aux data stream */
void
IntMutexContext::store(IntTagNode& sv, Stream& aux)
{
    // Storage format is:
    //   tag is flag
    //     0=not this process, 1=this process
    //   aux is
    //     2 words for string lengths
    //     2 strings (name, info)
    // Storing just an owner flag avoids the need to name processes.
    // If one process contains a variable containing a lock owned by
    // another one, the other one will claim the lock if he still
    // rightfully owns it; if he doesn't, it's probably better to
    // disown the lock.
    if (IntVMSaveContext* vsc = IntVMSaveContext::getCurrentInstance()) {
        sv.tag   = Tag_Mutex;
        sv.value = (mtx->getOwner() != 0 && mtx->getOwner() == vsc->getCurrentProcess());

        char header[8];
        storeUint32(header+0, mtx->getName().size());
        storeUint32(header+4, mtx->getNote().size());
        aux.writeT(header, sizeof(header));
        aux.writeT(mtx->getName().data(), mtx->getName().size());
        aux.writeT(mtx->getNote().data(), mtx->getNote().size());
    } else {
        throw IntError::notSerializable();
    }
}

/** Load implementation.
    \param flag Flag, from value of tag node.
    \param aux Aux stream */
IntMutexContext*
IntMutexContext::load(uint32_t flag, Stream& aux)
{
    // Find process owner
    IntExecutionContext* owner;
    if ((flag & 1) != 0) {
        // Owned by this process
        IntVMLoadContext* vlc = IntVMLoadContext::getCurrentInstance();
        if (vlc == 0 || vlc->getCurrentProcess() == 0) {
            return 0;
        }
        owner = vlc->getCurrentProcess();
    } else {
        owner = 0;
    }

    // Load data
    char header[8];
    string_t name, note;
    aux.readT(header, sizeof(header));
    getString(aux, name, getUint32(header+0));
    getString(aux, note, getUint32(header+4));

    // Find existing lock
    uint32_t slot = mutex_names.addMaybe(name);
    if (slot < mutex_objects.size() && mutex_objects[slot] != 0) {
        // This mutex already exists. Is it compatible?
        if (owner != 0) {
            if (mutex_objects[slot]->getOwner() != 0 && mutex_objects[slot]->getOwner() != owner) {
                throw IntError("Incompatible locks");
            }
            mutex_objects[slot]->setOwner(owner);
        }
        return new IntMutexContext(mutex_objects[slot]->addReference());
    } else {
        // Make new mutex
        return new IntMutexContext(IntMutex::createMutex(slot, note, owner));
    }
}

/******************************* Interface *******************************/

/** Implementation of Lock(). This must be implemented separately
    using the full interface (not the simplified one) because it needs
    an execution context. */
namespace {
    class IntLock : public IntCallableValue {
     public:
        IntLock();
        ~IntLock();

        // IntCallableValue:
        IntValue* call(IntExecutionContext& exc, IntArgBlock& args);
        bool isProcedureCall();

        // IntValue:
        string_t toString(bool readable);
        void store(IntTagNode& sv, Stream& aux);
        IntLock* clone();
    };
}

inline
IntLock::IntLock()
{ }

IntLock::~IntLock()
{ }

// IntCallableValue:
IntValue*
IntLock::call(IntExecutionContext& exc, IntArgBlock& args)
{
    args.checkArgumentCount(1, 2);

    string_t name, note;
    if (!checkStringArg(name, args.getNext()) || name.empty()) {
        throw IntError("Expecting lock name");
    }
    checkStringArg(note, args.getNext());

    return new IntMutexContext(IntMutex::create(strUCase(name), note, &exc));
}

bool
IntLock::isProcedureCall()
{
    return false;
}

// IntValue:
string_t
IntLock::toString(bool /*readable*/)
{
    return "Lock";
}

void
IntLock::store(IntTagNode& /*sv*/, Stream& /*aux*/)
{
    throw IntError::notSerializable();
}

IntLock*
IntLock::clone()
{
    return new IntLock();
}

/** Get lock information.
    - first arg: lock name
    - second arg, optional: type of information:
      - 0: true/false whether lock is there
      - 1: name of process holding the lock
      - 2: note */
static IntValue*
IFGetLockInfo(IntArgBlock& args)
{
    args.checkArgumentCount(1, 2);

    // Parse args
    string_t name;
    int32_t option = 0;
    if (!checkStringArg(name, args.getNext()))
        return 0;
    checkIntArg(option, args.getNext(), 0, 2);

    // Action
    IntMutex* mtx = IntMutex::query(strUCase(name));
    switch (option) {
     case 0:
        return makeBoolValue(mtx != 0);
     case 1:
        if (mtx != 0) {
            IntExecutionContext* exc = mtx->getOwner();
            if (exc != 0) {
                return makeStringValue(exc->getName());
            }
        }
        break;
     case 2:
        if (mtx != 0) {
            return makeStringValue(mtx->getNote());
        }
        break;
    }
    return 0;
}

/** Initialize interpreter mutex interface. */
void
initInterpreterMutexInterface()
{
    defineGlobalValue("LOCK", new IntLock());
    defineGlobalValue("GETLOCKINFO", new IntSimpleIndexableValue(IFGetLockInfo, 0));
}
